# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Python implementation of R-NaD (https://arxiv.org/pdf/2206.15378.pdf)."""

import enum
import functools
from typing import Any, Callable, Sequence, Tuple

import chex
import haiku as hk
import jax
from jax import lax
from jax import numpy as jnp
from jax import tree_util as tree
import numpy as np
import optax

from open_spiel.python import policy as policy_lib
import pyspiel


# Some handy aliases.
# Since most of these are just aliases for a "bag of tensors", the goal
# is to improve the documentation, and not to actually enforce correctness
# through pytype.
Params = chex.ArrayTree


class EntropySchedule:
  """An increasing list of steps where the regularisation network is updated.

  Example
    EntropySchedule([3, 5, 10], [2, 4, 1])
    =>   [0, 3, 6, 11, 16, 21, 26, 10]
          | 3 x2 |      5 x4     | 10 x1
  """

  def __init__(self, *, sizes: Sequence[int], repeats: Sequence[int]):
    """Constructs a schedule of entropy iterations.

    Args:
      sizes: the list of iteration sizes.
      repeats: the list, parallel to sizes, with the number of times for each
        size from `sizes` to repeat.
    """
    try:
      if len(repeats) != len(sizes):
        raise ValueError("`repeats` must be parallel to `sizes`.")
      if not sizes:
        raise ValueError("`sizes` and `repeats` must not be empty.")
      if any([(repeat <= 0) for repeat in repeats]):
        raise ValueError("All repeat values must be strictly positive")
      if repeats[-1] != 1:
        raise ValueError("The last value in `repeats` must be equal to 1, "
                         "ince the last iteration size is repeated forever.")
    except ValueError as e:
      raise ValueError(
          f"Entropy iteration schedule: repeats ({repeats}) and sizes"
          f" ({sizes})."
      ) from e

    schedule = [0]
    for size, repeat in zip(sizes, repeats):
      schedule.extend([schedule[-1] + (i + 1) * size for i in range(repeat)])

    self.schedule = np.array(schedule, dtype=np.int32)

  def __call__(self, learner_step: int) -> Tuple[float, bool]:
    """Entropy scheduling parameters for a given `learner_step`.

    Args:
      learner_step: The current learning step.

    Returns:
      alpha: The mixing weight (from [0, 1]) of the previous policy with
        the one before for computing the intrinsic reward.
      update_target_net: A boolean indicator for updating the target network
        with the current network.
    """

    # The complexity below is because at some point we might go past
    # the explicit schedule, and then we'd need to just use the last step
    # in the schedule and apply the logic of
    # ((learner_step - last_step) % last_iteration) == 0)

    # The schedule might look like this:
    # X----X-------X--X--X--X--------X
    # learner_step | might be here ^    |
    # or there     ^                    |
    # or even past the schedule         ^

    # We need to deal with two cases below.
    # Instead of going for the complicated conditional, let's just
    # compute both and then do the A * s + B * (1 - s) with s being a bool
    # selector between A and B.

    # 1. assume learner_step is past the schedule,
    #    ie schedule[-1] <= learner_step.
    last_size = self.schedule[-1] - self.schedule[-2]
    last_start = self.schedule[-1] + (
        learner_step - self.schedule[-1]) // last_size * last_size
    # 2. assume learner_step is within the schedule.
    start = jnp.amax(self.schedule * (self.schedule <= learner_step))
    finish = jnp.amin(
        self.schedule * (learner_step < self.schedule),
        initial=self.schedule[-1],
        where=(learner_step < self.schedule))
    size = finish - start

    # Now select between the two.
    beyond = (self.schedule[-1] <= learner_step)  # Are we past the schedule?
    iteration_start = (last_start * beyond + start * (1 - beyond))
    iteration_size = (last_size * beyond + size * (1 - beyond))

    update_target_net = jnp.logical_and(
        learner_step > 0, jnp.sum(learner_step == iteration_start))
    alpha = jnp.minimum(
        (2.0 * (learner_step - iteration_start)) / iteration_size, 1.0)

    return alpha, update_target_net  # pytype: disable=bad-return-type  # jax-types


@chex.dataclass(frozen=True)
class FineTuning:
  """Fine tuning options, aka policy post-processing.

  Even when fully trained, the resulting softmax-based policy may put
  a small probability mass on bad actions. This results in an agent
  waiting for the opponent (itself in self-play) to commit an error.

  To address that the policy is post-processed using:
  - thresholding: any action with probability smaller than self.threshold
    is simply removed from the policy.
  - discretization: the probability values are rounded to the closest
    multiple of 1/self.discretization.

  The post-processing is used on the learner, and thus must be jit-friendly.
  """
  # The learner step after which the policy post processing (aka finetuning)
  # will be enabled when learning. A strictly negative value is equivalent
  # to infinity, ie disables finetuning completely.
  from_learner_steps: int = -1
  # All policy probabilities below `threshold` are zeroed out. Thresholding
  # is disabled if this value is non-positive.
  policy_threshold: float = 0.03
  # Rounds the policy probabilities to the "closest"
  # multiple of 1/`self.discretization`.
  # Discretization is disabled for non-positive values.
  policy_discretization: int = 32

  def __call__(self, policy: chex.Array, mask: chex.Array,
               learner_steps: int) -> chex.Array:
    """A configurable fine tuning of a policy."""
    chex.assert_equal_shape((policy, mask))
    do_finetune = jnp.logical_and(self.from_learner_steps >= 0,
                                  learner_steps > self.from_learner_steps)

    return jnp.where(do_finetune, self.post_process_policy(policy, mask),
                     policy)

  def post_process_policy(
      self,
      policy: chex.Array,
      mask: chex.Array,
  ) -> chex.Array:
    """Unconditionally post process a given masked policy."""
    chex.assert_equal_shape((policy, mask))
    policy = self._threshold(policy, mask)
    policy = self._discretize(policy)
    return policy

  def _threshold(self, policy: chex.Array, mask: chex.Array) -> chex.Array:
    """Remove from the support the actions 'a' where policy(a) < threshold."""
    chex.assert_equal_shape((policy, mask))
    if self.policy_threshold <= 0:
      return policy

    mask = mask * (
        # Values over the threshold.
        (policy >= self.policy_threshold) +
        # Degenerate case is when policy is less than threshold *everywhere*.
        # In that case we just keep the policy as-is.
        (jnp.max(policy, axis=-1, keepdims=True) < self.policy_threshold))
    return mask * policy / jnp.sum(mask * policy, axis=-1, keepdims=True)

  def _discretize(self, policy: chex.Array) -> chex.Array:
    """Round all action probabilities to a multiple of 1/self.discretize."""
    if self.policy_discretization <= 0:
      return policy

    # The unbatched/single policy case:
    if len(policy.shape) == 1:
      return self._discretize_single(policy)

    # policy may be [B, A] or [T, B, A], etc. Thus add hk.BatchApply.
    dims = len(policy.shape) - 1

    # TODO(author18): avoid mixing vmap and BatchApply since the two could
    # be folded into either a single BatchApply or a sequence of vmaps, but
    # not the mix.
    vmapped = jax.vmap(self._discretize_single)
    policy = hk.BatchApply(vmapped, num_dims=dims)(policy)

    return policy

  def _discretize_single(self, mu: chex.Array) -> chex.Array:
    """A version of self._discretize but for the unbatched data."""
    # TODO(author18): try to merge _discretize and _discretize_single
    # into one function that handles both batched and unbatched cases.
    if len(mu.shape) == 2:
      mu_ = jnp.squeeze(mu, axis=0)
    else:
      mu_ = mu
    n_actions = mu_.shape[-1]
    roundup = jnp.ceil(mu_ * self.policy_discretization).astype(jnp.int32)
    result = jnp.zeros_like(mu_)
    order = jnp.argsort(-mu_)  # Indices of descending order.
    weight_left = self.policy_discretization

    def f_disc(i, order, roundup, weight_left, result):
      x = jnp.minimum(roundup[order[i]], weight_left)
      result = jax.numpy.where(weight_left >= 0, result.at[order[i]].add(x),
                               result)
      weight_left -= x
      return i + 1, order, roundup, weight_left, result

    def f_scan_scan(carry, x):
      i, order, roundup, weight_left, result = carry
      i_next, order_next, roundup_next, weight_left_next, result_next = f_disc(
          i, order, roundup, weight_left, result)
      carry_next = (i_next, order_next, roundup_next, weight_left_next,
                    result_next)
      return carry_next, x

    (_, _, _, weight_left_next, result_next), _ = jax.lax.scan(
        f_scan_scan,
        init=(jnp.asarray(0), order, roundup, weight_left, result),
        xs=None,
        length=n_actions)

    result_next = jnp.where(weight_left_next > 0,
                            result_next.at[order[0]].add(weight_left_next),
                            result_next)
    if len(mu.shape) == 2:
      result_next = jnp.expand_dims(result_next, axis=0)
    return result_next / self.policy_discretization


def _legal_policy(logits: chex.Array, legal_actions: chex.Array) -> chex.Array:
  """A soft-max policy that respects legal_actions."""
  chex.assert_equal_shape((logits, legal_actions))
  # Fiddle a bit to make sure we don't generate NaNs or Inf in the middle.
  l_min = logits.min(axis=-1, keepdims=True)
  logits = jnp.where(legal_actions, logits, l_min)
  logits -= logits.max(axis=-1, keepdims=True)
  logits *= legal_actions
  exp_logits = jnp.where(legal_actions, jnp.exp(logits),
                         0)  # Illegal actions become 0.
  exp_logits_sum = jnp.sum(exp_logits, axis=-1, keepdims=True)
  return exp_logits / exp_logits_sum


def legal_log_policy(logits: chex.Array,
                     legal_actions: chex.Array) -> chex.Array:
  """Return the log of the policy on legal action, 0 on illegal action."""
  chex.assert_equal_shape((logits, legal_actions))
  # logits_masked has illegal actions set to -inf.
  logits_masked = logits + jnp.log(legal_actions)
  max_legal_logit = logits_masked.max(axis=-1, keepdims=True)
  logits_masked = logits_masked - max_legal_logit
  # exp_logits_masked is 0 for illegal actions.
  exp_logits_masked = jnp.exp(logits_masked)

  baseline = jnp.log(jnp.sum(exp_logits_masked, axis=-1, keepdims=True))
  # Subtract baseline from logits. We do not simply return
  #     logits_masked - baseline
  # because that has -inf for illegal actions, or
  #     legal_actions * (logits_masked - baseline)
  # because that leads to 0 * -inf == nan for illegal actions.
  log_policy = jnp.multiply(legal_actions,
                            (logits - max_legal_logit - baseline))
  return log_policy


def _player_others(player_ids: chex.Array, valid: chex.Array,
                   player: int) -> chex.Array:
  """A vector of 1 for the current player and -1 for others.

  Args:
    player_ids: Tensor [...] containing player ids (0 <= player_id < N).
    valid: Tensor [...] containing whether these states are valid.
    player: The player id as int.

  Returns:
    player_other: is 1 for the current player and -1 for others [..., 1].
  """
  chex.assert_equal_shape((player_ids, valid))
  current_player_tensor = (player_ids == player).astype(jnp.int32)  # pytype: disable=attribute-error  # numpy-scalars

  res = 2 * current_player_tensor - 1
  res = res * valid
  return jnp.expand_dims(res, axis=-1)


def _policy_ratio(pi: chex.Array, mu: chex.Array, actions_oh: chex.Array,
                  valid: chex.Array) -> chex.Array:
  """Returns a ratio of policy pi/mu when selecting action a.

  By convention, this ratio is 1 on non valid states
  Args:
    pi: the policy of shape [..., A].
    mu: the sampling policy of shape [..., A].
    actions_oh: a one-hot encoding of the current actions of shape [..., A].
    valid: 0 if the state is not valid and else 1 of shape [...].

  Returns:
    pi/mu on valid states and 1 otherwise. The shape is the same
    as pi, mu or actions_oh but without the last dimension A.
  """
  chex.assert_equal_shape((pi, mu, actions_oh))
  chex.assert_shape((valid,), actions_oh.shape[:-1])

  def _select_action_prob(pi):
    return (jnp.sum(actions_oh * pi, axis=-1, keepdims=False) * valid +
            (1 - valid))

  pi_actions_prob = _select_action_prob(pi)
  mu_actions_prob = _select_action_prob(mu)
  return pi_actions_prob / mu_actions_prob


def _where(pred: chex.Array, true_data: chex.ArrayTree,
           false_data: chex.ArrayTree) -> chex.ArrayTree:
  """Similar to jax.where but treats `pred` as a broadcastable prefix."""

  def _where_one(t, f):
    chex.assert_equal_rank((t, f))
    # Expand the dimensions of pred if true_data and false_data are higher rank.
    p = jnp.reshape(pred, pred.shape + (1,) * (len(t.shape) - len(pred.shape)))
    return jnp.where(p, t, f)

  return tree.tree_map(_where_one, true_data, false_data)


def _has_played(valid: chex.Array, player_id: chex.Array,
                player: int) -> chex.Array:
  """Compute a mask of states which have a next state in the sequence."""
  chex.assert_equal_shape((valid, player_id))

  def _loop_has_played(carry, x):
    valid, player_id = x
    chex.assert_equal_shape((valid, player_id))

    our_res = jnp.ones_like(player_id)
    opp_res = carry
    reset_res = jnp.zeros_like(carry)

    our_carry = carry
    opp_carry = carry
    reset_carry = jnp.zeros_like(player_id)

    # pyformat: disable
    return _where(valid, _where((player_id == player),
                                (our_carry, our_res),
                                (opp_carry, opp_res)),
                  (reset_carry, reset_res))
    # pyformat: enable

  _, result = lax.scan(
      f=_loop_has_played,
      init=jnp.zeros_like(player_id[-1]),
      xs=(valid, player_id),
      reverse=True)
  return result


# V-Trace
#
# Custom implementation of VTrace to handle trajectories having a mix of
# different player steps. The standard rlax.vtrace can't be applied here
# out of the box because a trajectory could look like '121211221122'.


def v_trace(
    v: chex.Array,
    valid: chex.Array,
    player_id: chex.Array,
    acting_policy: chex.Array,
    merged_policy: chex.Array,
    merged_log_policy: chex.Array,
    player_others: chex.Array,
    actions_oh: chex.Array,
    reward: chex.Array,
    player: int,
    # Scalars below.
    eta: float,
    lambda_: float,
    c: float,
    rho: float,
) -> Tuple[Any, Any, Any]:
  """Custom VTrace for trajectories with a mix of different player steps."""
  gamma = 1.0

  has_played = _has_played(valid, player_id, player)

  policy_ratio = _policy_ratio(merged_policy, acting_policy, actions_oh, valid)
  inv_mu = _policy_ratio(
      jnp.ones_like(merged_policy), acting_policy, actions_oh, valid)

  eta_reg_entropy = (-eta *
                     jnp.sum(merged_policy * merged_log_policy, axis=-1) *
                     jnp.squeeze(player_others, axis=-1))
  eta_log_policy = -eta * merged_log_policy * player_others

  @chex.dataclass(frozen=True)
  class LoopVTraceCarry:
    """The carry of the v-trace scan loop."""
    reward: chex.Array
    # The cumulated reward until the end of the episode. Uncorrected (v-trace).
    # Gamma discounted and includes eta_reg_entropy.
    reward_uncorrected: chex.Array
    next_value: chex.Array
    next_v_target: chex.Array
    importance_sampling: chex.Array

  init_state_v_trace = LoopVTraceCarry(
      reward=jnp.zeros_like(reward[-1]),
      reward_uncorrected=jnp.zeros_like(reward[-1]),
      next_value=jnp.zeros_like(v[-1]),
      next_v_target=jnp.zeros_like(v[-1]),
      importance_sampling=jnp.ones_like(policy_ratio[-1]))

  def _loop_v_trace(carry: LoopVTraceCarry, x) -> Tuple[LoopVTraceCarry, Any]:
    (cs, player_id, v, reward, eta_reg_entropy, valid, inv_mu, actions_oh,
     eta_log_policy) = x

    reward_uncorrected = (
        reward + gamma * carry.reward_uncorrected + eta_reg_entropy)
    discounted_reward = reward + gamma * carry.reward

    # V-target:
    our_v_target = (
        v + jnp.expand_dims(
            jnp.minimum(rho, cs * carry.importance_sampling), axis=-1) *
        (jnp.expand_dims(reward_uncorrected, axis=-1) +
         gamma * carry.next_value - v) + lambda_ * jnp.expand_dims(
             jnp.minimum(c, cs * carry.importance_sampling), axis=-1) * gamma *
        (carry.next_v_target - carry.next_value))

    opp_v_target = jnp.zeros_like(our_v_target)
    reset_v_target = jnp.zeros_like(our_v_target)

    # Learning output:
    our_learning_output = (
        v +  # value
        eta_log_policy +  # regularisation
        actions_oh * jnp.expand_dims(inv_mu, axis=-1) *
        (jnp.expand_dims(discounted_reward, axis=-1) + gamma * jnp.expand_dims(
            carry.importance_sampling, axis=-1) * carry.next_v_target - v))

    opp_learning_output = jnp.zeros_like(our_learning_output)
    reset_learning_output = jnp.zeros_like(our_learning_output)

    # State carry:
    our_carry = LoopVTraceCarry(
        reward=jnp.zeros_like(carry.reward),
        next_value=v,
        next_v_target=our_v_target,
        reward_uncorrected=jnp.zeros_like(carry.reward_uncorrected),
        importance_sampling=jnp.ones_like(carry.importance_sampling))
    opp_carry = LoopVTraceCarry(
        reward=eta_reg_entropy + cs * discounted_reward,
        reward_uncorrected=reward_uncorrected,
        next_value=gamma * carry.next_value,
        next_v_target=gamma * carry.next_v_target,
        importance_sampling=cs * carry.importance_sampling)
    reset_carry = init_state_v_trace

    # Invalid turn: init_state_v_trace and (zero target, learning_output)
    # pyformat: disable
    return _where(valid,  # pytype: disable=bad-return-type  # numpy-scalars
                  _where((player_id == player),
                         (our_carry, (our_v_target, our_learning_output)),
                         (opp_carry, (opp_v_target, opp_learning_output))),
                  (reset_carry, (reset_v_target, reset_learning_output)))
    # pyformat: enable

  _, (v_target, learning_output) = lax.scan(
      f=_loop_v_trace,
      init=init_state_v_trace,
      xs=(policy_ratio, player_id, v, reward, eta_reg_entropy, valid, inv_mu,
          actions_oh, eta_log_policy),
      reverse=True)

  return v_target, has_played, learning_output


def get_loss_v(v_list: Sequence[chex.Array],
               v_target_list: Sequence[chex.Array],
               mask_list: Sequence[chex.Array]) -> chex.Array:
  """Define the loss function for the critic."""
  chex.assert_trees_all_equal_shapes(v_list, v_target_list)
  # v_list and v_target_list come with a degenerate trailing dimension,
  # which mask_list tensors do not have.
  chex.assert_shape(mask_list, v_list[0].shape[:-1])
  loss_v_list = []
  for (v_n, v_target, mask) in zip(v_list, v_target_list, mask_list):
    assert v_n.shape[0] == v_target.shape[0]

    loss_v = jnp.expand_dims(
        mask, axis=-1) * (v_n - lax.stop_gradient(v_target))**2
    normalization = jnp.sum(mask)
    loss_v = jnp.sum(loss_v) / (normalization + (normalization == 0.0))

    loss_v_list.append(loss_v)
  return sum(loss_v_list)


def apply_force_with_threshold(decision_outputs: chex.Array, force: chex.Array,
                               threshold: float,
                               threshold_center: chex.Array) -> chex.Array:
  """Apply the force with below a given threshold."""
  chex.assert_equal_shape((decision_outputs, force, threshold_center))
  can_decrease = decision_outputs - threshold_center > -threshold
  can_increase = decision_outputs - threshold_center < threshold
  force_negative = jnp.minimum(force, 0.0)
  force_positive = jnp.maximum(force, 0.0)
  clipped_force = can_decrease * force_negative + can_increase * force_positive
  return decision_outputs * lax.stop_gradient(clipped_force)


def renormalize(loss: chex.Array, mask: chex.Array) -> chex.Array:
  """The `normalization` is the number of steps over which loss is computed."""
  chex.assert_equal_shape((loss, mask))
  loss = jnp.sum(loss * mask)
  normalization = jnp.sum(mask)
  return loss / (normalization + (normalization == 0.0))


def get_loss_nerd(logit_list: Sequence[chex.Array],
                  policy_list: Sequence[chex.Array],
                  q_vr_list: Sequence[chex.Array],
                  valid: chex.Array,
                  player_ids: Sequence[chex.Array],
                  legal_actions: chex.Array,
                  importance_sampling_correction: Sequence[chex.Array],
                  clip: float = 100,
                  threshold: float = 2) -> chex.Array:
  """Define the nerd loss."""
  assert isinstance(importance_sampling_correction, list)
  loss_pi_list = []
  for k, (logit_pi, pi, q_vr, is_c) in enumerate(
      zip(logit_list, policy_list, q_vr_list, importance_sampling_correction)):
    assert logit_pi.shape[0] == q_vr.shape[0]
    # loss policy
    adv_pi = q_vr - jnp.sum(pi * q_vr, axis=-1, keepdims=True)
    adv_pi = is_c * adv_pi  # importance sampling correction
    adv_pi = jnp.clip(adv_pi, a_min=-clip, a_max=clip)
    adv_pi = lax.stop_gradient(adv_pi)

    logits = logit_pi - jnp.mean(
        logit_pi * legal_actions, axis=-1, keepdims=True)

    threshold_center = jnp.zeros_like(logits)

    nerd_loss = jnp.sum(
        legal_actions *
        apply_force_with_threshold(logits, adv_pi, threshold, threshold_center),
        axis=-1)
    nerd_loss = -renormalize(nerd_loss, valid * (player_ids == k))
    loss_pi_list.append(nerd_loss)
  return sum(loss_pi_list)


@chex.dataclass(frozen=True)
class AdamConfig:
  """Adam optimizer related params."""
  b1: float = 0.0
  b2: float = 0.999
  eps: float = 10e-8


@chex.dataclass(frozen=True)
class NerdConfig:
  """Nerd related params."""
  beta: float = 2.0
  clip: float = 10_000


class StateRepresentation(str, enum.Enum):
  INFO_SET = "info_set"
  OBSERVATION = "observation"


@chex.dataclass(frozen=True)
class RNaDConfig:
  """Configuration parameters for the RNaDSolver."""
  # The game parameter string including its name and parameters.
  game_name: str
  # The games longer than this value are truncated. Must be strictly positive.
  trajectory_max: int = 10

  # The content of the EnvStep.obs tensor.
  state_representation: StateRepresentation = StateRepresentation.INFO_SET

  # Network configuration.
  policy_network_layers: Sequence[int] = (256, 256)

  # The batch size to use when learning/improving parameters.
  batch_size: int = 256
  # The learning rate for `params`.
  learning_rate: float = 0.00005
  # The config related to the ADAM optimizer used for updating `params`.
  adam: AdamConfig = AdamConfig()
  # All gradients values are clipped to [-clip_gradient, clip_gradient].
  clip_gradient: float = 10_000
  # The "speed" at which `params_target` is following `params`.
  target_network_avg: float = 0.001

  # RNaD algorithm configuration.
  # Entropy schedule configuration. See EntropySchedule class documentation.
  entropy_schedule_repeats: Sequence[int] = (1,)
  entropy_schedule_size: Sequence[int] = (20_000,)
  # The weight of the reward regularisation term in RNaD.
  eta_reward_transform: float = 0.2
  nerd: NerdConfig = NerdConfig()
  c_vtrace: float = 1.0

  # Options related to fine tuning of the agent.
  finetune: FineTuning = FineTuning()

  # The seed that fully controls the randomness.
  seed: int = 42


@chex.dataclass(frozen=True)
class EnvStep:
  """Holds the tensor data representing the current game state."""
  # Indicates whether the state is a valid one or just a padding. Shape: [...]
  # The terminal state being the first one to be marked !valid.
  # All other tensors in EnvStep contain data, but only for valid timesteps.
  # Once !valid the data needs to be ignored, since it's a duplicate of
  # some other previous state.
  # The rewards is the only exception that contains reward values
  # in the terminal state, which is marked !valid.
  # TODO(author16): This is a confusion point and would need to be clarified.
  valid: chex.Array = ()  # pytype: disable=annotation-type-mismatch  # numpy-scalars
  # The single tensor representing the state observation. Shape: [..., ??]
  obs: chex.Array = ()  # pytype: disable=annotation-type-mismatch  # numpy-scalars
  # The legal actions mask for the current player. Shape: [..., A]
  legal: chex.Array = ()  # pytype: disable=annotation-type-mismatch  # numpy-scalars
  # The current player id as an int. Shape: [...]
  player_id: chex.Array = ()  # pytype: disable=annotation-type-mismatch  # numpy-scalars
  # The rewards of all the players. Shape: [..., P]
  rewards: chex.Array = ()  # pytype: disable=annotation-type-mismatch  # numpy-scalars


@chex.dataclass(frozen=True)
class ActorStep:
  """The actor step tensor summary."""
  # The action (as one-hot) of the current player. Shape: [..., A]
  action_oh: chex.Array = ()  # pytype: disable=annotation-type-mismatch  # numpy-scalars
  # The policy of the current player. Shape: [..., A]
  policy: chex.Array = ()  # pytype: disable=annotation-type-mismatch  # numpy-scalars
  # The rewards of all the players. Shape: [..., P]
  # Note - these are rewards obtained *after* the actor step, and thus
  # these are the same as EnvStep.rewards visible before the *next* step.
  rewards: chex.Array = ()  # pytype: disable=annotation-type-mismatch  # numpy-scalars


@chex.dataclass(frozen=True)
class TimeStep:
  """The tensor data for one game transition (env_step, actor_step)."""
  env: EnvStep = EnvStep()
  actor: ActorStep = ActorStep()


Optimizer = Callable[[Params, Params], Params]  # (params, grads) -> params


def optax_optimizer(
    params: chex.ArrayTree,
    init_and_update: optax.GradientTransformation) -> Optimizer:
  """Creates a parameterized function that represents an optimizer."""
  init_fn, update_fn = init_and_update

  @chex.dataclass
  class OptaxOptimizer:
    """A jax-friendly representation of an optimizer state with the update."""
    state: chex.Array

    def __call__(self, params: Params, grads: Params) -> Params:
      updates, self.state = update_fn(grads, self.state)  # pytype: disable=annotation-type-mismatch  # numpy-scalars
      return optax.apply_updates(params, updates)

  return OptaxOptimizer(state=init_fn(params))


class RNaDSolver(policy_lib.Policy):
  """Implements a solver for the R-NaD Algorithm.

  See https://arxiv.org/abs/2206.15378.

  Define all networks. Derive losses & learning steps. Initialize the game
  state and algorithmic variables.
  """

  def __init__(self, config: RNaDConfig):
    self.config = config

    # Learner and actor step counters.
    self.learner_steps = 0
    self.actor_steps = 0

    self.init()

  def init(self):
    """Initialize the network and losses."""
    # The random facilities for jax and numpy.
    self._rngkey = jax.random.PRNGKey(self.config.seed)
    self._np_rng = np.random.RandomState(self.config.seed)
    # TODO(author16): serialize both above to get the fully deterministic behaviour.

    # Create a game and an example of a state.
    self._game = pyspiel.load_game(self.config.game_name)
    self._ex_state = self._play_chance(self._game.new_initial_state())

    # The network.
    def network(
        env_step: EnvStep
    ) -> Tuple[chex.Array, chex.Array, chex.Array, chex.Array]:
      mlp_torso = hk.nets.MLP(
          self.config.policy_network_layers, activate_final=True
      )
      torso = mlp_torso(env_step.obs)

      mlp_policy_head = hk.nets.MLP([self._game.num_distinct_actions()])
      logit = mlp_policy_head(torso)

      mlp_policy_value = hk.nets.MLP([1])
      v = mlp_policy_value(torso)

      pi = _legal_policy(logit, env_step.legal)
      log_pi = legal_log_policy(logit, env_step.legal)
      return pi, v, log_pi, logit

    self.network = hk.without_apply_rng(hk.transform(network))

    # The machinery related to updating parameters/learner.
    self._entropy_schedule = EntropySchedule(
        sizes=self.config.entropy_schedule_size,
        repeats=self.config.entropy_schedule_repeats)
    self._loss_and_grad = jax.value_and_grad(self.loss, has_aux=False)

    # Create initial parameters.
    env_step = self._state_as_env_step(self._ex_state)
    key = self._next_rng_key()  # Make sure to use the same key for all.
    self.params = self.network.init(key, env_step)
    self.params_target = self.network.init(key, env_step)
    self.params_prev = self.network.init(key, env_step)
    self.params_prev_ = self.network.init(key, env_step)

    # Parameter optimizers.
    self.optimizer = optax_optimizer(
        self.params,
        optax.chain(
            optax.scale_by_adam(
                eps_root=0.0,
                **self.config.adam,
            ), optax.scale(-self.config.learning_rate),
            optax.clip(self.config.clip_gradient)))
    self.optimizer_target = optax_optimizer(
        self.params_target, optax.sgd(self.config.target_network_avg))

  def loss(self, params: Params, params_target: Params, params_prev: Params,
           params_prev_: Params, ts: TimeStep, alpha: float,
           learner_steps: int) -> float:
    rollout = jax.vmap(self.network.apply, (None, 0), 0)
    pi, v, log_pi, logit = rollout(params, ts.env)

    policy_pprocessed = self.config.finetune(pi, ts.env.legal, learner_steps)

    _, v_target, _, _ = rollout(params_target, ts.env)
    _, _, log_pi_prev, _ = rollout(params_prev, ts.env)
    _, _, log_pi_prev_, _ = rollout(params_prev_, ts.env)
    # This line creates the reward transform log(pi(a|x)/pi_reg(a|x)).
    # For the stability reasons, reward changes smoothly between iterations.
    # The mixing between old and new reward transform is a convex combination
    # parametrised by alpha.
    log_policy_reg = log_pi - (alpha * log_pi_prev + (1 - alpha) * log_pi_prev_)

    v_target_list, has_played_list, v_trace_policy_target_list = [], [], []
    for player in range(self._game.num_players()):
      reward = ts.actor.rewards[:, :, player]  # [T, B, Player]
      v_target_, has_played, policy_target_ = v_trace(
          v_target,
          ts.env.valid,
          ts.env.player_id,
          ts.actor.policy,
          policy_pprocessed,
          log_policy_reg,
          _player_others(ts.env.player_id, ts.env.valid, player),
          ts.actor.action_oh,
          reward,
          player,
          lambda_=1.0,
          c=self.config.c_vtrace,
          rho=np.inf,
          eta=self.config.eta_reward_transform)
      v_target_list.append(v_target_)
      has_played_list.append(has_played)
      v_trace_policy_target_list.append(policy_target_)
    loss_v = get_loss_v([v] * self._game.num_players(), v_target_list,
                        has_played_list)

    is_vector = jnp.expand_dims(jnp.ones_like(ts.env.valid), axis=-1)
    importance_sampling_correction = [is_vector] * self._game.num_players()
    # Uses v-trace to define q-values for Nerd
    loss_nerd = get_loss_nerd(
        [logit] * self._game.num_players(), [pi] * self._game.num_players(),
        v_trace_policy_target_list,
        ts.env.valid,
        ts.env.player_id,
        ts.env.legal,
        importance_sampling_correction,
        clip=self.config.nerd.clip,
        threshold=self.config.nerd.beta)
    return loss_v + loss_nerd  # pytype: disable=bad-return-type  # numpy-scalars

  @functools.partial(jax.jit, static_argnums=(0,))
  def update_parameters(
      self,
      params: Params,
      params_target: Params,
      params_prev: Params,
      params_prev_: Params,
      optimizer: Optimizer,
      optimizer_target: Optimizer,
      timestep: TimeStep,
      alpha: float,
      learner_steps: int,
      update_target_net: bool):
    """A jitted pure-functional part of the `step`."""
    loss_val, grad = self._loss_and_grad(params, params_target, params_prev,
                                         params_prev_, timestep, alpha,
                                         learner_steps)
    # Update `params`` using the computed gradient.
    params = optimizer(params, grad)
    # Update `params_target` towards `params`.
    params_target = optimizer_target(
        params_target, tree.tree_map(lambda a, b: a - b, params_target, params))

    # Rolls forward the prev and prev_ params if update_target_net is 1.
    # pyformat: disable
    params_prev, params_prev_ = jax.lax.cond(
        update_target_net,
        lambda: (params_target, params_prev),
        lambda: (params_prev, params_prev_))
    # pyformat: enable

    logs = {
        "loss": loss_val,
    }
    return (params, params_target, params_prev, params_prev_, optimizer,
            optimizer_target), logs

  def __getstate__(self):
    """To serialize the agent."""
    return dict(
        # RNaD config.
        config=self.config,

        # Learner and actor step counters.
        learner_steps=self.learner_steps,
        actor_steps=self.actor_steps,

        # The randomness keys.
        np_rng=self._np_rng.get_state(),
        rngkey=self._rngkey,

        # Network params.
        params=self.params,
        params_target=self.params_target,
        params_prev=self.params_prev,
        params_prev_=self.params_prev_,
        # Optimizer state.
        optimizer=self.optimizer.state,  # pytype: disable=attribute-error  # always-use-return-annotations
        optimizer_target=self.optimizer_target.state,  # pytype: disable=attribute-error  # always-use-return-annotations
    )

  def __setstate__(self, state):
    """To deserialize the agent."""
    # RNaD config.
    self.config = state["config"]

    self.init()

    # Learner and actor step counters.
    self.learner_steps = state["learner_steps"]
    self.actor_steps = state["actor_steps"]

    # The randomness keys.
    self._np_rng.set_state(state["np_rng"])
    self._rngkey = state["rngkey"]

    # Network params.
    self.params = state["params"]
    self.params_target = state["params_target"]
    self.params_prev = state["params_prev"]
    self.params_prev_ = state["params_prev_"]
    # Optimizer state.
    self.optimizer.state = state["optimizer"]
    self.optimizer_target.state = state["optimizer_target"]

  def step(self):
    """One step of the algorithm, that plays the game and improves params."""
    timestep = self.collect_batch_trajectory()
    alpha, update_target_net = self._entropy_schedule(self.learner_steps)
    (self.params, self.params_target, self.params_prev, self.params_prev_,
     self.optimizer, self.optimizer_target), logs = self.update_parameters(
         self.params, self.params_target, self.params_prev, self.params_prev_,
         self.optimizer, self.optimizer_target, timestep, alpha,
         self.learner_steps, update_target_net)
    self.learner_steps += 1
    logs.update({
        "actor_steps": self.actor_steps,
        "learner_steps": self.learner_steps,
    })
    return logs

  def _next_rng_key(self) -> chex.PRNGKey:
    """Get the next rng subkey from class rngkey.

    Must *not* be called from under a jitted function!

    Returns:
      A fresh rng_key.
    """
    self._rngkey, subkey = jax.random.split(self._rngkey)
    return subkey

  def _state_as_env_step(self, state: pyspiel.State) -> EnvStep:
    # A terminal state must be communicated to players, however since
    # it's a terminal state things like the state_representation or
    # the set of legal actions are meaningless and only needed
    # for the sake of creating well a defined trajectory tensor.
    # Therefore the code below:
    # - extracts the rewards
    # - if the state is terminal, uses a dummy other state for other fields.
    rewards = np.array(state.returns(), dtype=np.float64)

    valid = not state.is_terminal()
    if not valid:
      state = self._ex_state

    if self.config.state_representation == StateRepresentation.OBSERVATION:
      obs = state.observation_tensor()
    elif self.config.state_representation == StateRepresentation.INFO_SET:
      obs = state.information_state_tensor()
    else:
      raise ValueError(
          f"Invalid StateRepresentation: {self.config.state_representation}.")

    # TODO(author16): clarify the story around rewards and valid.
    return EnvStep(
        obs=np.array(obs, dtype=np.float64),
        legal=np.array(state.legal_actions_mask(), dtype=np.int8),
        player_id=np.array(state.current_player(), dtype=np.float64),
        valid=np.array(valid, dtype=np.float64),
        rewards=rewards)

  def action_probabilities(self,
                           state: pyspiel.State,
                           player_id: Any = None):
    """Returns action probabilities dict for a single batch."""
    env_step = self._batch_of_states_as_env_step([state])
    probs = self._network_jit_apply_and_post_process(
        self.params_target, env_step)
    probs = jax.device_get(probs[0])  # Squeeze out the 1-element batch.
    return {
        action: probs[action]
        for action, valid in enumerate(jax.device_get(env_step.legal[0]))
        if valid
    }

  @functools.partial(jax.jit, static_argnums=(0,))
  def _network_jit_apply_and_post_process(
      self, params: Params, env_step: EnvStep) -> chex.Array:
    pi, _, _, _ = self.network.apply(params, env_step)
    pi = self.config.finetune.post_process_policy(pi, env_step.legal)
    return pi

  # TODO(author16): jit actor_step.
  def actor_step(self, env_step: EnvStep):
    pi, _, _, _ = self.network.apply(self.params, env_step)
    pi = np.asarray(pi).astype("float64")
    # TODO(author18): is this policy normalization really needed?
    pi = pi / np.sum(pi, axis=-1, keepdims=True)

    action = np.apply_along_axis(
        lambda x: self._np_rng.choice(range(pi.shape[1]), p=x), axis=-1, arr=pi)
    # TODO(author16): reapply the legal actions mask to bullet-proof sampling.
    action_oh = np.zeros(pi.shape, dtype="float64")
    action_oh[range(pi.shape[0]), action] = 1.0

    actor_step = ActorStep(policy=pi, action_oh=action_oh, rewards=())  # pytype: disable=wrong-arg-types  # numpy-scalars

    return action, actor_step

  def collect_batch_trajectory(self) -> TimeStep:
    states = [
        self._play_chance(self._game.new_initial_state())
        for _ in range(self.config.batch_size)
    ]
    timesteps = []

    env_step = self._batch_of_states_as_env_step(states)
    for _ in range(self.config.trajectory_max):
      prev_env_step = env_step
      a, actor_step = self.actor_step(env_step)

      states = self._batch_of_states_apply_action(states, a)
      env_step = self._batch_of_states_as_env_step(states)
      timesteps.append(
          TimeStep(
              env=prev_env_step,
              actor=ActorStep(
                  action_oh=actor_step.action_oh,
                  policy=actor_step.policy,
                  rewards=env_step.rewards),
          ))
    # Concatenate all the timesteps together to form a single rollout [T, B, ..]
    return jax.tree_util.tree_map(lambda *xs: np.stack(xs, axis=0), *timesteps)

  def _batch_of_states_as_env_step(self,
                                   states: Sequence[pyspiel.State]) -> EnvStep:
    envs = [self._state_as_env_step(state) for state in states]
    return jax.tree_util.tree_map(lambda *e: np.stack(e, axis=0), *envs)

  def _batch_of_states_apply_action(
      self, states: Sequence[pyspiel.State],
      actions: chex.Array) -> Sequence[pyspiel.State]:
    """Apply a batch of `actions` to a parallel list of `states`."""
    for state, action in zip(states, list(actions)):
      if not state.is_terminal():
        self.actor_steps += 1
        state.apply_action(action)
        self._play_chance(state)
    return states

  def _play_chance(self, state: pyspiel.State) -> pyspiel.State:
    """Plays the chance nodes until we end up at another type of node.

    Args:
      state: to be updated until it does not correspond to a chance node.
    Returns:
      The same input state object, but updated. The state is returned
      only for convenience, to allow chaining function calls.
    """
    while state.is_chance_node():
      chance_outcome, chance_proba = zip(*state.chance_outcomes())
      action = self._np_rng.choice(chance_outcome, p=chance_proba)
      state.apply_action(action)
    return state



# Copyright 2021 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
r"""Two BlueChip bridge bots agains simplest open_spiel (take the first possible action).

The bot_cmd FLAG should contain a command-line to launch an external bot, e.g.
`Wbridge5 Autoconnect {port}`.

"""
# pylint: enable=line-too-long

import os
import pickle
import re
import socket
import subprocess
import time

from absl import app
from absl import flags
import haiku as hk
import jax
import numpy as np

from open_spiel.python.bots import bluechip_bridge
import pyspiel

FLAGS = flags.FLAGS
flags.DEFINE_float("timeout_secs", 60, "Seconds to wait for bot to respond")
flags.DEFINE_integer("rng_seed", 1234, "Seed to use to generate hands")
flags.DEFINE_integer("num_deals", 10, "How many deals to play")
flags.DEFINE_integer("sleep", 0, "How many seconds to wait before next action")
flags.DEFINE_string("params_path", ".",
                    "directory path for trained model params-snapshot.pkl")
flags.DEFINE_string(
    "bot_cmd", None,
    "Command to launch the external bot; must include {port} which will be "
    "replaced by the port number to attach to.")

# Make the network.
NUM_ACTIONS = 38
MIN_ACTION = 52


def net_fn(x):
  """Haiku module for our network."""
  net = hk.Sequential([
      hk.Linear(1024),
      jax.nn.relu,
      hk.Linear(1024),
      jax.nn.relu,
      hk.Linear(1024),
      jax.nn.relu,
      hk.Linear(1024),
      jax.nn.relu,
      hk.Linear(NUM_ACTIONS),
      jax.nn.log_softmax,
  ])
  return net(x)


def load_model():
  net = hk.without_apply_rng(hk.transform(net_fn))
  params = pickle.load(
      open(os.path.join(FLAGS.params_path, "params-snapshot.pkl"), "rb"))
  return net, params


def ai_action(state, net, params):
  observation = np.array(state.observation_tensor(), np.float32)
  policy = np.exp(net.apply(params, observation))
  probs_actions = [(p, a + MIN_ACTION) for a, p in enumerate(policy)]
  pred = max(probs_actions)[1]
  return pred


def _run_once(state, bots, net, params):
  """Plays bots with each other, returns terminal utility for each player."""
  for bot in bots:
    bot.restart()
  while not state.is_terminal():
    if state.is_chance_node():
      outcomes, probs = zip(*state.chance_outcomes())
      state.apply_action(np.random.choice(outcomes, p=probs))
    else:
      if FLAGS.sleep:
        time.sleep(FLAGS.sleep)  # wait for the human to see how it goes
      if state.current_player() % 2 == 1:
        # Have simplest play for now
        action = state.legal_actions()[0]
        if action > 51:
          # TODO(ed2k) extend beyond just bidding
          action = ai_action(state, net, params)
        state.apply_action(action)
      else:
        result = bots[state.current_player() // 2].step(state)
        state.apply_action(result)
  return state


def main(argv):
  if len(argv) > 1:
    raise app.UsageError("Too many command-line arguments.")
  game = pyspiel.load_game("bridge(use_double_dummy_result=false)")
  net, params = load_model()
  bots = [
      bluechip_bridge.BlueChipBridgeBot(game, 0, controller_factory),
      bluechip_bridge.BlueChipBridgeBot(game, 2, controller_factory)
  ]

  results = []

  for i_deal in range(FLAGS.num_deals):
    state = _run_once(game.new_initial_state(), bots, net, params)
    print("Deal #{}; final state:\n{}".format(i_deal, state))
    results.append(state.returns())

  stats = np.array(results)
  mean = np.mean(stats, axis=0)
  stderr = np.std(stats, axis=0, ddof=1) / np.sqrt(FLAGS.num_deals)
  print(u"Absolute score: {:+.1f}\u00b1{:.1f}".format(mean[0], stderr[0]))
  print(u"Relative score: {:+.1f}\u00b1{:.1f}".format(mean[1], stderr[1]))


def controller_factory():
  """Implements bluechip_bridge.BlueChipBridgeBot."""
  client = _WBridge5Client(FLAGS.bot_cmd)
  client.start()
  return client


class _WBridge5Client(object):
  """Manages the connection to a WBridge5 bot."""

  def __init__(self, command):
    self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    self.sock.bind(("", 0))
    self.port = self.sock.getsockname()[1]
    self.sock.listen(1)
    self.process = None
    self.command = command.format(port=self.port)

  def start(self):
    if self.process is not None:
      self.process.kill()
    self.process = subprocess.Popen(self.command.split(" "))
    self.conn, self.addr = self.sock.accept()

  def read_line(self):
    line = ""
    while True:
      self.conn.settimeout(FLAGS.timeout_secs)
      data = self.conn.recv(1024)
      if not data:
        raise EOFError("Connection closed")
      line += data.decode("ascii")
      if line.endswith("\n"):
        return re.sub(r"\s+", " ", line).strip()

  def send_line(self, line):
    self.conn.send((line + "\r\n").encode("ascii"))

  def terminate(self):
    self.process.kill()
    self.process = None


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.algorithms.action_value.py."""

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import action_value
import pyspiel


class ActionValuesTest(parameterized.TestCase):

  @parameterized.parameters([["kuhn_poker", 2], ["kuhn_poker", 3],
                             ["leduc_poker", 2]])
  def test_runs_with_uniform_policies(self, game_name, num_players):
    game = pyspiel.load_game(game_name, {"players": num_players})
    calc = action_value.TreeWalkCalculator(game)
    uniform_policy = policy.TabularPolicy(game)
    calc.compute_all_states_action_values([uniform_policy] * num_players)

  def test_kuhn_poker_always_pass_p0(self):
    game = pyspiel.load_game("kuhn_poker")
    calc = action_value.TreeWalkCalculator(game)
    uniform_policy = policy.TabularPolicy(game)
    always_pass_policy = policy.FirstActionPolicy(game).to_tabular()
    returned_values = calc([always_pass_policy, uniform_policy],
                           always_pass_policy)
    root_node_values = calc.get_root_node_values(
        [always_pass_policy, uniform_policy])
    self.assertTrue(
        np.allclose(root_node_values, returned_values.root_node_values))

    # Action 0 == Pass. Action 1 == Bet
    # Some values are 0 because the states are not reached, thus the expected
    # value of that node is undefined.
    np.testing.assert_array_almost_equal(
        np.asarray([
            # Player 0 states
            [-1.0, -0.5],    # '0'
            [-1.0, -2.0],    # '0pb'
            [-0.5, 0.5],     # '1'
            [-1.0, 0.0],     # '1pb'
            [0.0, 1.5],      # '2'
            [-1.0, 2.0],     # '2pb'
            # Player 1 states
            [0.0, 1.0],      # '1p'
            [0, 0],          # Unreachable
            [1.0, 1.0],      # '2p'
            [0, 0],          # Unreachable
            [-1.0, 1.0],     # '0p'
            [0, 0],          # Unreachable
        ]), returned_values.action_values)

    np.testing.assert_array_almost_equal(
        np.asarray([
            # Player 0 states
            1 / 3,  # '0'
            1 / 6,  # '0pb'
            1 / 3,  # '1'
            1 / 6,  # '1pb'
            1 / 3,  # '2'
            1 / 6,  # '2pb'
            # Player 1 states
            1 / 3,  # '1p'
            0.0,  # '1b': zero because player 0 always play pass
            1 / 3,  # 2p'
            0.0,  # '2b': zero because player 0 always play pass
            1 / 3,  # '0p'
            0.0,  # '0b':  zero because player 0 always play pass
        ]),
        returned_values.counterfactual_reach_probs)

    # The reach probabilities are always one, even though we have player 0
    # who only plays pass, because the unreachable nodes for player 0 are
    # terminal nodes: e.g.  'x x b b p' has a player 0 reach of 0, but it is
    # a terminal node, thus it does not appear in the tabular policy
    # states.
    np.testing.assert_array_equal(
        [1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0],
        returned_values.player_reach_probs)

    np.testing.assert_array_almost_equal(
        np.asarray([
            np.array([-1/3, -1/6]),
            np.array([-1/6, -1/3]),
            np.array([-1/6, 1/6]),
            np.array([-1/6, 0.]),
            np.array([0., 0.5]),
            np.array([-1/6, 1/3]),
            np.array([0., 1/3]),
            np.array([0., 0.]),
            np.array([1/3, 1/3]),
            np.array([0., 0.]),
            np.array([-1/3, 1/3]),
            np.array([0., 0.])
        ]), returned_values.sum_cfr_reach_by_action_value)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tabular Q-learning agent."""

import collections
import numpy as np

from open_spiel.python import rl_agent
from open_spiel.python import rl_tools


def valuedict():
  # The default factory is called without arguments to produce a new value when
  # a key is not present, in __getitem__ only. This value is added to the dict,
  # so modifying it will modify the dict.
  return collections.defaultdict(float)


class QLearner(rl_agent.AbstractAgent):
  """Tabular Q-Learning agent.

  See open_spiel/python/examples/tic_tac_toe_qlearner.py for an usage example.
  """

  def __init__(self,
               player_id,
               num_actions,
               step_size=0.1,
               epsilon_schedule=rl_tools.ConstantSchedule(0.2),
               discount_factor=1.0,
               centralized=False):
    """Initialize the Q-Learning agent."""
    self._player_id = player_id
    self._num_actions = num_actions
    self._step_size = step_size
    self._epsilon_schedule = epsilon_schedule
    self._epsilon = epsilon_schedule.value
    self._discount_factor = discount_factor
    self._centralized = centralized
    self._q_values = collections.defaultdict(valuedict)
    self._prev_info_state = None
    self._last_loss_value = None

  def _epsilon_greedy(self, info_state, legal_actions, epsilon):
    """Returns a valid epsilon-greedy action and valid action probs.

    If the agent has not been to `info_state`, a valid random action is chosen.

    Args:
      info_state: hashable representation of the information state.
      legal_actions: list of actions at `info_state`.
      epsilon: float, prob of taking an exploratory action.

    Returns:
      A valid epsilon-greedy action and valid action probabilities.
    """
    probs = np.zeros(self._num_actions)
    greedy_q = max([self._q_values[info_state][a] for a in legal_actions])
    greedy_actions = [
        a for a in legal_actions if self._q_values[info_state][a] == greedy_q
    ]
    probs[legal_actions] = epsilon / len(legal_actions)
    probs[greedy_actions] += (1 - epsilon) / len(greedy_actions)
    action = np.random.choice(range(self._num_actions), p=probs)
    return action, probs

  def _get_action_probs(self, info_state, legal_actions, epsilon):
    """Returns a selected action and the probabilities of legal actions.

    To be overwritten by subclasses that implement other action selection
    methods.

    Args:
      info_state: hashable representation of the information state.
      legal_actions: list of actions at `info_state`.
      epsilon: float: current value of the epsilon schedule or 0 in case
        evaluation. QLearner uses it as the exploration parameter in
        epsilon-greedy, but subclasses are free to interpret in different ways
        (e.g. as temperature in softmax).
    """
    return self._epsilon_greedy(info_state, legal_actions, epsilon)

  def step(self, time_step, is_evaluation=False):
    """Returns the action to be taken and updates the Q-values if needed.

    Args:
      time_step: an instance of rl_environment.TimeStep.
      is_evaluation: bool, whether this is a training or evaluation call.

    Returns:
      A `rl_agent.StepOutput` containing the action probs and chosen action.
    """
    if self._centralized:
      info_state = str(time_step.observations["info_state"])
    else:
      info_state = str(time_step.observations["info_state"][self._player_id])
    legal_actions = time_step.observations["legal_actions"][self._player_id]

    # Prevent undefined errors if this agent never plays until terminal step
    action, probs = None, None

    # Act step: don't act at terminal states.
    if not time_step.last():
      epsilon = 0.0 if is_evaluation else self._epsilon
      action, probs = self._get_action_probs(info_state, legal_actions, epsilon)

    # Learn step: don't learn during evaluation or at first agent steps.
    if self._prev_info_state and not is_evaluation:
      target = time_step.rewards[self._player_id]
      if not time_step.last():  # Q values are zero for terminal.
        target += self._discount_factor * max(
            [self._q_values[info_state][a] for a in legal_actions])

      prev_q_value = self._q_values[self._prev_info_state][self._prev_action]
      self._last_loss_value = target - prev_q_value
      self._q_values[self._prev_info_state][self._prev_action] += (
          self._step_size * self._last_loss_value)

      # Decay epsilon, if necessary.
      self._epsilon = self._epsilon_schedule.step()

      if time_step.last():  # prepare for the next episode.
        self._prev_info_state = None
        return

    # Don't mess up with the state during evaluation.
    if not is_evaluation:
      self._prev_info_state = info_state
      self._prev_action = action
    return rl_agent.StepOutput(action=action, probs=probs)

  @property
  def loss(self):
    return self._last_loss_value



# Copyright (c) 2019-2021 The Pybind Development Team. All rights reserved.
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""Tests for absl pybind11 casters."""

import array
import datetime

from absl.testing import absltest
from absl.testing import parameterized
from dateutil import tz
import numpy as np

from pybind11_abseil.tests import absl_example


class AbslTimeTest(parameterized.TestCase):
  SECONDS_IN_DAY = 24 * 60 * 60
  POSITIVE_SECS = 3 * SECONDS_IN_DAY + 2.5
  NEGATIVE_SECS = -3 * SECONDS_IN_DAY + 2.5
  TEST_DATETIME = datetime.datetime(2000, 1, 2, 3, 4, 5, int(5e5))
  # Linter error relevant for pytz only.
  # pylint: disable=g-tzinfo-replace
  TEST_DATETIME_UTC = TEST_DATETIME.replace(tzinfo=tz.tzutc())
  # pylint: enable=g-tzinfo-replace
  TEST_DATE = datetime.date(2000, 1, 2)

  def test_return_positive_duration(self):
    duration = absl_example.make_duration(self.POSITIVE_SECS)
    self.assertEqual(duration.days, 3)
    self.assertEqual(duration.seconds, 2)
    self.assertEqual(duration.microseconds, 5e5)

  def test_return_negative_duration(self):
    duration = absl_example.make_duration(self.NEGATIVE_SECS)
    self.assertEqual(duration.days, -3)
    self.assertEqual(duration.seconds, 2)
    self.assertEqual(duration.microseconds, 5e5)

  def test_pass_positive_duration(self):
    duration = datetime.timedelta(seconds=self.POSITIVE_SECS)
    self.assertTrue(absl_example.check_duration(duration, self.POSITIVE_SECS))

  def test_pass_negative_duration(self):
    duration = datetime.timedelta(seconds=self.NEGATIVE_SECS)
    self.assertTrue(absl_example.check_duration(duration, self.NEGATIVE_SECS))

  def test_pass_float_duration(self):
    self.assertTrue(
        absl_example.check_duration(self.POSITIVE_SECS, self.POSITIVE_SECS))

  def test_pass_integer_duration(self):
    self.assertTrue(
        absl_example.check_duration(self.SECONDS_IN_DAY, self.SECONDS_IN_DAY))

  def test_duration_integer_overflow(self):
    duration = 2**129
    with self.assertRaises(RuntimeError):
      absl_example.check_duration(duration, duration)

  def test_return_datetime(self):
    secs = self.TEST_DATETIME.timestamp()
    local_tz = tz.gettz()
    # pylint: disable=g-tzinfo-datetime
    # Warning about tzinfo applies to pytz, but we are using dateutil.tz
    expected_datetime = datetime.datetime(
        year=self.TEST_DATETIME.year,
        month=self.TEST_DATETIME.month,
        day=self.TEST_DATETIME.day,
        hour=self.TEST_DATETIME.hour,
        minute=self.TEST_DATETIME.minute,
        second=self.TEST_DATETIME.second,
        microsecond=self.TEST_DATETIME.microsecond,
        tzinfo=local_tz)
    # pylint: enable=g-tzinfo-datetime
    # datetime handling code will set the local timezone on C++ times for want
    # of a better alternative
    self.assertEqual(expected_datetime, absl_example.make_datetime(secs))

  def test_pass_date(self):
    secs = datetime.datetime(self.TEST_DATE.year, self.TEST_DATE.month,
                             self.TEST_DATE.day).timestamp()
    self.assertTrue(absl_example.check_datetime(self.TEST_DATE, secs))

  def test_pass_datetime(self):
    secs = self.TEST_DATETIME.timestamp()
    self.assertTrue(absl_example.check_datetime(self.TEST_DATETIME, secs))

  def test_pass_datetime_with_timezone(self):
    pacific_tz = tz.gettz('America/Los_Angeles')
    # pylint: disable=g-tzinfo-datetime
    # Warning about tzinfo applies to pytz, but we are using dateutil.tz
    dt_with_tz = datetime.datetime(
        year=2020, month=2, day=1, hour=20, tzinfo=pacific_tz)
    # pylint: enable=g-tzinfo-datetime
    secs = dt_with_tz.timestamp()
    self.assertTrue(absl_example.check_datetime(dt_with_tz, secs))

  def test_pass_datetime_dst_with_timezone(self):
    pacific_tz = tz.gettz('America/Los_Angeles')
    # pylint: disable=g-tzinfo-datetime
    dst_end = datetime.datetime(2020, 11, 1, 2, 0, 0, tzinfo=pacific_tz)
    # pylint: enable=g-tzinfo-datetime
    secs = dst_end.timestamp()
    self.assertTrue(absl_example.check_datetime(dst_end, secs))

  def test_pass_datetime_dst(self):
    dst_end = datetime.datetime(2020, 11, 1, 2, 0, 0)
    secs = dst_end.timestamp()
    self.assertTrue(absl_example.check_datetime(dst_end, secs))

  @parameterized.named_parameters(('before', -1), ('flip', 0), ('after', 1))
  def test_dst_datetime_from_timestamp(self, offs):
    secs_flip = 1604224799  # 2020-11-01T02:00:00-08:00
    secs = secs_flip + offs
    time_utc = datetime.datetime.fromtimestamp(secs, tz.tzutc())
    time_local_aware = time_utc.astimezone(tz.gettz())
    time_local_naive = time_local_aware.replace(tzinfo=None)
    for time in (time_utc, time_local_aware, time_local_naive):
      self.assertTrue(absl_example.check_datetime(time, secs))

  def test_pass_datetime_pre_unix_epoch(self):
    dt = datetime.datetime(1969, 7, 16, 10, 56, 7, microsecond=140)
    secs = dt.timestamp()
    self.assertTrue(absl_example.check_datetime(dt, secs))

  def test_return_civilsecond(self):
    # We need to use a timezone aware datetime here, otherwise
    # datetime.timestamp converts to localtime. UTC is chosen as the convention
    # in the test cases.
    truncated = self.TEST_DATETIME.replace(microsecond=0)
    self.assertEqual(
        truncated,
        absl_example.make_civilsecond(self.TEST_DATETIME_UTC.timestamp()))

  def test_pass_datetime_as_civilsecond(self):
    truncated = self.TEST_DATETIME_UTC.replace(microsecond=0)
    self.assertTrue(
        absl_example.check_civilsecond(self.TEST_DATETIME,
                                       truncated.timestamp()))

  def test_return_civilminute(self):
    truncated = self.TEST_DATETIME.replace(second=0, microsecond=0)
    self.assertEqual(
        truncated,
        absl_example.make_civilminute(self.TEST_DATETIME_UTC.timestamp()))

  def test_pass_datetime_as_civilminute(self):
    truncated = self.TEST_DATETIME_UTC.replace(second=0, microsecond=0)
    self.assertTrue(
        absl_example.check_civilminute(self.TEST_DATETIME,
                                       truncated.timestamp()))

  def test_return_civilhour(self):
    truncated = self.TEST_DATETIME.replace(minute=0, second=0, microsecond=0)
    self.assertEqual(
        truncated,
        absl_example.make_civilhour(self.TEST_DATETIME_UTC.timestamp()))

  def test_pass_datetime_as_civilhour(self):
    truncated = self.TEST_DATETIME_UTC.replace(
        minute=0, second=0, microsecond=0)
    self.assertTrue(
        absl_example.check_civilhour(self.TEST_DATETIME, truncated.timestamp()))

  def test_return_civilday(self):
    truncated = self.TEST_DATETIME.replace(
        hour=0, minute=0, second=0, microsecond=0)
    self.assertEqual(
        truncated,
        absl_example.make_civilday(self.TEST_DATETIME_UTC.timestamp()))

  def test_pass_datetime_as_civilday(self):
    truncated = self.TEST_DATETIME_UTC.replace(
        hour=0, minute=0, second=0, microsecond=0)
    self.assertTrue(
        absl_example.check_civilday(self.TEST_DATETIME, truncated.timestamp()))

  def test_return_civilmonth(self):
    truncated = self.TEST_DATETIME.replace(
        day=1, hour=0, minute=0, second=0, microsecond=0)
    self.assertEqual(
        truncated,
        absl_example.make_civilmonth(self.TEST_DATETIME_UTC.timestamp()))

  def test_pass_datetime_as_civilmonth(self):
    truncated = self.TEST_DATETIME_UTC.replace(
        day=1, hour=0, minute=0, second=0, microsecond=0)
    self.assertTrue(
        absl_example.check_civilmonth(self.TEST_DATETIME,
                                      truncated.timestamp()))

  def test_return_civilyear(self):
    truncated = self.TEST_DATETIME.replace(
        month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
    self.assertEqual(
        truncated,
        absl_example.make_civilyear(self.TEST_DATETIME_UTC.timestamp()))

  def test_pass_datetime_as_civilyear(self):
    truncated = self.TEST_DATETIME_UTC.replace(
        month=1, day=1, hour=0, minute=0, second=0, microsecond=0)
    self.assertTrue(
        absl_example.check_civilyear(self.TEST_DATETIME, truncated.timestamp()))

  def test_timezone(self):
    expected_timezone = 'Fixed/UTC+02:00:00'
    timezone = absl_example.roundtrip_timezone(expected_timezone)
    self.assertEqual(expected_timezone, timezone)
    timezone = absl_example.roundtrip_timezone(2 * 60 * 60)
    self.assertEqual(expected_timezone, timezone)
    with self.assertRaises(TypeError):
      absl_example.roundtrip_timezone('Not a timezone')


def make_read_only_numpy_array():
  values = np.zeros(5, dtype=np.int32)
  values.flags.writeable = False
  return values


def make_srided_numpy_array(stride):
  return np.zeros(10, dtype=np.int32)[::stride]


class AbslNumericSpanTest(parameterized.TestCase):

  # The check_span* functions use signed ints. Arrays with different numeric
  # types and all non-array types (eg native sequence types) require converting.
  CONVERTED_NUMERIC_LISTS = (
      ('array_wrong_dtype', array.array('b', [9, 8, 7])),
      ('numpy_wrong_dtype', np.array([7, 8, 9], dtype=np.uint16)),
      ('tuple', (1, 2, 3)),
      ('list', [4, 5, 6]),
  )

  NOT_CONVERTED_NUMERIC_LISTS = (
      ('array_matching_dtype', array.array('i', [9, 8, 7])),
      ('numpy_matching_dtype', np.array([7, 8, 9], dtype=np.int32)),
  )

  NUMERIC_LISTS = CONVERTED_NUMERIC_LISTS + NOT_CONVERTED_NUMERIC_LISTS

  def test_return_span(self):
    values = [1, 2, 3, 4]
    container = absl_example.VectorContainer()
    self.assertSequenceEqual(container.make_span(values), values)

  @parameterized.named_parameters(*NUMERIC_LISTS)
  def test_pass_span_from(self, values):
    # Pass values twice- one will be converted to a span, the other to a vector
    # (which is known to work), and then they will be compared.
    self.assertTrue(absl_example.check_span(values, values))

  @parameterized.named_parameters(*NUMERIC_LISTS)
  def test_caster_copy_from(self, values):
    self.assertTrue(absl_example.check_span_caster_copy(values, values))

  @parameterized.named_parameters(*NOT_CONVERTED_NUMERIC_LISTS)
  def test_pass_span_no_convert_from(self, values):
    self.assertTrue(absl_example.check_span_no_convert(values, values))

  @parameterized.named_parameters(*CONVERTED_NUMERIC_LISTS)
  def test_pass_span_no_convert_fails_from(self, values):
    with self.assertRaises(TypeError):
      absl_example.check_span_no_convert(values, values)

  @parameterized.named_parameters(
      ('string', 'test'), ('int', 5), ('object', absl_example.ObjectForSpan(3)),
      ('list_of_objects', [absl_example.ObjectForSpan(3)]))
  def test_pass_span_fails_from(self, values):
    with self.assertRaises(TypeError):
      absl_example.check_span(values, values)

  def test_fill_span_from_numpy(self):
    values = np.zeros(5, dtype=np.int32)
    absl_example.fill_span(42, values)
    for e in values:
      self.assertEqual(e, 42)

  @parameterized.named_parameters(
      ('float_numpy', np.zeros(5, dtype=np.float)),
      ('two_d_numpy', np.zeros(
          (5, 5), dtype=np.int32)), ('read_only', make_read_only_numpy_array()),
      ('strided_skip', make_srided_numpy_array(2)),
      ('strided_reverse', make_srided_numpy_array(-1)),
      ('non_supported_type', np.zeros(5, dtype=np.unicode_)),
      ('native_list', [0] * 5))
  def test_fill_span_fails_from(self, values):
    with self.assertRaises(TypeError):
      absl_example.fill_span(42, values)


def make_native_list_of_objects():
  return [absl_example.ObjectForSpan(3), absl_example.ObjectForSpan(5)]


def make_opaque_list_of_objects():
  return absl_example.ObjectVector(make_native_list_of_objects())


def make_list_of_objects_parameters():
  return (('native_list', make_native_list_of_objects()),
          ('opaque_list', make_opaque_list_of_objects()))


class AbslObjectSpanTest(parameterized.TestCase):

  @parameterized.named_parameters(*make_list_of_objects_parameters())
  def test_pass_object_pointers_span_from(self, objects):
    self.assertEqual(absl_example.sum_object_pointers_span(objects), 8)

  @parameterized.named_parameters(*make_list_of_objects_parameters())
  def test_pass_object_span_from(self, objects):
    self.assertEqual(absl_example.sum_object_span(objects), 8)

  def test_pass_object_span_no_convert_from_opaque_list(self):
    absl_example.sum_object_span_no_convert(make_opaque_list_of_objects())

  def test_pass_object_span_no_convert_fails_from_native_list(self):
    # Native list requires conversion.
    with self.assertRaises(TypeError):
      absl_example.sum_object_span_no_convert(make_native_list_of_objects())

  @parameterized.named_parameters(*make_list_of_objects_parameters())
  def test_fill_object_pointers_span_from(self, objects):
    absl_example.fill_object_pointers_span(42, objects)
    for obj in objects:
      self.assertEqual(obj.value, 42)

  def test_fill_object_span_no_convert_from_opaque_list(self):
    objects = make_opaque_list_of_objects()
    absl_example.fill_object_span(42, objects)
    for obj in objects:
      self.assertEqual(obj.value, 42)

  def test_fill_object_span_no_convert_fails_from_native_list(self):
    objects = make_native_list_of_objects()
    # Native list requires conversion.
    with self.assertRaises(TypeError):
      absl_example.fill_object_span(objects)


class AbslStringViewTest(absltest.TestCase):
  TEST_STRING = 'test string!'

  def test_return_view(self):
    container = absl_example.StringContainer()
    self.assertSequenceEqual(
        container.make_string_view(self.TEST_STRING), self.TEST_STRING)

  def test_pass_string_view(self):
    self.assertTrue(
        absl_example.check_string_view(self.TEST_STRING, self.TEST_STRING))


class AbslCordTest(absltest.TestCase):
  TEST_STRING = 'absl_Cord'
  TEST_BYTES = b'absl_Cord'

  def test_return_absl_cord(self):
    self.assertSequenceEqual(
        absl_example.return_absl_cord(self.TEST_STRING), self.TEST_STRING)
    self.assertSequenceEqual(
        absl_example.return_absl_cord(self.TEST_BYTES), self.TEST_STRING)

  def test_pass_absl_cord(self):
    self.assertTrue(
        absl_example.check_absl_cord(self.TEST_STRING, self.TEST_STRING))
    self.assertFalse(
        absl_example.check_absl_cord(self.TEST_STRING, '12345'))
    self.assertTrue(
        absl_example.check_absl_cord(self.TEST_BYTES, self.TEST_STRING))


class AbslFlatHashMapTest(absltest.TestCase):

  def test_return_map(self):
    keys_and_values = [(1, 2), (3, 4), (5, 6)]
    expected = dict(keys_and_values)
    self.assertEqual(expected, absl_example.make_map(keys_and_values))

  def test_pass_map(self):
    expected = [(10, 20), (30, 40)]
    self.assertTrue(absl_example.check_map(dict(expected), expected))


class AbslFlatHashSetTest(absltest.TestCase):

  def test_return_set(self):
    values = [1, 3, 7, 5]
    expected = set(values)
    self.assertEqual(expected, absl_example.make_set(values))

  def test_pass_set(self):
    expected = [10, 20, 30, 40]
    self.assertTrue(absl_example.check_set(set(expected), expected))


class AbslBTreeMapTest(absltest.TestCase):

  def test_return_map(self):
    keys_and_values = [(1, 2), (3, 4), (5, 6)]
    expected = dict(keys_and_values)
    self.assertEqual(expected, absl_example.make_btree_map(keys_and_values))

  def test_pass_map(self):
    expected = [(10, 20), (30, 40)]
    self.assertTrue(absl_example.check_btree_map(dict(expected), expected))


class AbslOptionalTest(absltest.TestCase):

  def test_pass_default_nullopt(self):
    self.assertTrue(absl_example.check_optional())

  def test_pass_value(self):
    self.assertTrue(absl_example.check_optional(5, True, 5))

  def test_pass_none(self):
    self.assertTrue(absl_example.check_optional(None, False))

  def test_return_value(self):
    self.assertEqual(absl_example.make_optional(5), 5)

  def test_return_none(self):
    self.assertIsNone(absl_example.make_optional())


class AbslVariantTest(absltest.TestCase):

  def test_variant(self):
    assert absl_example.VariantToInt(absl_example.A(3)) == 3
    assert absl_example.VariantToInt(absl_example.B(5)) == 5

    for identity_f, should_be_equal in [(absl_example.Identity, True),
                                        (absl_example.IdentityWithCopy, False)]:
      objs = [absl_example.A(3), absl_example.B(5)]
      vector = identity_f(objs)
      self.assertLen(vector, 2)
      self.assertIsInstance(vector[0], absl_example.A)
      self.assertEqual(vector[0].a, 3)
      self.assertIsInstance(vector[1], absl_example.B)
      self.assertEqual(vector[1].b, 5)
      if should_be_equal:
        self.assertEqual(objs, vector)
      else:
        self.assertNotEqual(objs, vector)


if __name__ == '__main__':
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pickle

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python import test_utils
import pyspiel

# All games with kSampledStochastic chance mode.
SPIEL_SAMPLED_STOCHASTIC_GAMES_LIST = [
    g for g in pyspiel.registered_games() if g.default_loadable and
    g.chance_mode == pyspiel.GameType.ChanceMode.SAMPLED_STOCHASTIC
]
assert len(SPIEL_SAMPLED_STOCHASTIC_GAMES_LIST) >= 2

# We only do 2 runs as this is slow.
NUM_RUNS = 2


class SampledStochasticGamesTest(parameterized.TestCase):

  @parameterized.parameters(*SPIEL_SAMPLED_STOCHASTIC_GAMES_LIST)
  def test_stateful_game_serialization(self, game_info):
    game = pyspiel.load_game(game_info.short_name, {"rng_seed": 0})

    for seed in range(NUM_RUNS):
      # Mutate game's internal RNG state by doing a full playout.
      test_utils.random_playout(game.new_initial_state(), seed)
      deserialized_game = pickle.loads(pickle.dumps(game))

      # Make sure initial states are the same after game deserialization.
      state = test_utils.random_playout(game.new_initial_state(), seed)
      deserialized_state = test_utils.random_playout(
          deserialized_game.new_initial_state(), seed)
      self.assertEqual(str(state), str(deserialized_state))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Unit test for uniform random bot."""

import random

from absl.testing import absltest
from open_spiel.python.bots import uniform_random
import pyspiel


class BotTest(absltest.TestCase):

  def test_policy_is_uniform(self):
    game = pyspiel.load_game("leduc_poker")
    bots = [
        uniform_random.UniformRandomBot(0, random),
        uniform_random.UniformRandomBot(1, random)
    ]

    # deal each player a card
    state = game.new_initial_state()
    state.apply_action(2)
    state.apply_action(4)

    # p0 starts: uniform from [check, bet]
    policy, _ = bots[0].step_with_policy(state)
    self.assertCountEqual(policy, [(1, 0.5), (2, 0.5)])

    # Afte p0 bets, p1 chooses from [fold, call, raise]
    state.apply_action(2)
    policy, _ = bots[1].step_with_policy(state)
    self.assertCountEqual(policy, [(0, 1 / 3), (1, 1 / 3), (2, 1 / 3)])

  def test_no_legal_actions(self):
    game = pyspiel.load_game("kuhn_poker")
    bot = uniform_random.UniformRandomBot(0, random)
    state = game.new_initial_state()
    state.apply_action(2)  # deal
    state.apply_action(1)  # deal
    state.apply_action(1)  # bet
    state.apply_action(0)  # fold
    bot.restart_at(state)
    policy, action = bot.step_with_policy(state)
    self.assertEqual(policy, [])
    self.assertEqual(action, pyspiel.INVALID_ACTION)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for open_spiel.python.jax.opponent_shaping."""
import typing
from typing import Tuple
from absl.testing import absltest
from absl.testing import parameterized

import distrax
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np
from open_spiel.python import rl_environment
from open_spiel.python.jax.opponent_shaping import OpponentShapingAgent
import pyspiel

SEED = 24984617


def make_iterated_matrix_game(
    game: str, iterations=5, batch_size=8
) -> rl_environment.Environment:
  matrix_game = pyspiel.load_matrix_game(game)
  config = {'num_repetitions': iterations, 'batch_size': batch_size}
  game = pyspiel.create_repeated_game(matrix_game, config)
  env = rl_environment.Environment(game)
  return env


def make_agent_networks(
    num_actions: int,
) -> Tuple[hk.Transformed, hk.Transformed]:
  def policy(obs):
    logits = hk.nets.MLP(output_sizes=[8, 8, num_actions], with_bias=True)(obs)
    logits = jnp.nan_to_num(logits)
    return distrax.Categorical(logits=logits)

  def value_fn(obs):
    values = hk.nets.MLP(output_sizes=[8, 8, 1], with_bias=True)(obs)
    return values

  return hk.without_apply_rng(hk.transform(policy)), hk.without_apply_rng(
      hk.transform(value_fn)
  )


def run_agents(
    agents: typing.List[OpponentShapingAgent],
    env: rl_environment.Environment,
    num_steps=1000,
):
  time_step = env.reset()
  for _ in range(num_steps):
    actions = []
    for agent in agents:
      action, _ = agent.step(time_step)
      if action is not None:
        action = action.squeeze()
      actions.append(action)
    if time_step.last():
      time_step = env.reset()
    else:
      time_step = env.step(actions)
      time_step.observations['actions'] = np.array(actions)


class LolaPolicyGradientTest(parameterized.TestCase, absltest.TestCase):

  @parameterized.parameters(['matrix_pd'])
  def test_run_game(self, game_name):
    batch_size = 8
    iterations = 5
    env = make_iterated_matrix_game(
        game_name, batch_size=1, iterations=iterations
    )
    env.seed(SEED)
    key = jax.random.PRNGKey(SEED)
    num_actions = env.action_spec()['num_actions']
    policy_network, critic_network = make_agent_networks(
        num_actions=num_actions
    )

    # pylint: disable=g-complex-comprehension
    agents = [
        OpponentShapingAgent(
            player_id=i,
            opponent_ids=[1 - i],
            seed=key,
            correction_type='lola',
            env=env,
            n_lookaheads=1,
            info_state_size=env.observation_spec()['info_state'],
            num_actions=env.action_spec()['num_actions'],
            policy=policy_network,
            critic=critic_network,
            batch_size=batch_size,
            pi_learning_rate=0.005,
            critic_learning_rate=1.0,
            policy_update_interval=2,
            discount=0.96,
            use_jit=False,
        )
        for i in range(2)
    ]
    run_agents(agents=agents, env=env, num_steps=batch_size * 10)


class DicePolicyGradientTest(parameterized.TestCase, absltest.TestCase):

  @parameterized.parameters(['matrix_pd'])
  def test_run_game(self, game_name):
    batch_size = 8
    iterations = 5
    env = make_iterated_matrix_game(
        game_name, batch_size=1, iterations=iterations
    )
    env.seed(SEED)
    key = jax.random.PRNGKey(SEED)
    num_actions = env.action_spec()['num_actions']
    policy_network, critic_network = make_agent_networks(
        num_actions=num_actions
    )

    # pylint: disable=g-complex-comprehension
    agents = [
        OpponentShapingAgent(
            player_id=i,
            opponent_ids=[1 - i],
            seed=key,
            correction_type='dice',
            env=env,
            n_lookaheads=2,
            info_state_size=env.observation_spec()['info_state'],
            num_actions=env.action_spec()['num_actions'],
            policy=policy_network,
            critic=critic_network,
            batch_size=batch_size,
            pi_learning_rate=0.005,
            critic_learning_rate=1.0,
            policy_update_interval=2,
            discount=0.96,
            use_jit=False,
        )
        for i in range(2)
    ]
    run_agents(agents=agents, env=env, num_steps=batch_size * 10)


if __name__ == '__main__':
  np.random.seed(SEED)
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Munchausen Online Mirror Descent."""

from typing import Dict, List, Optional

import numpy as np

from open_spiel.python import policy as policy_lib
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import mirror_descent
import pyspiel


class ProjectedPolicyMunchausen(mirror_descent.ProjectedPolicy):
  """Project values on the policy simplex."""

  def __init__(
      self,
      game: pyspiel.Game,
      player_ids: List[int],
      state_value: value.ValueFunction,
      learning_rate: float,
      policy: policy_lib.Policy,
  ):
    """Initializes the projected policy.

    Args:
      game: The game to analyze.
      player_ids: list of player ids for which this policy applies; each should
        be in the range 0..game.num_players()-1.
      state_value: The state value to project.
      learning_rate: The learning rate.
      policy: The policy to project.
    """
    super().__init__(game, player_ids, state_value)
    self._learning_rate = learning_rate
    self._policy = policy

  def action_probabilities(self,
                           state: pyspiel.State,
                           player_id: Optional[int] = None) -> Dict[int, float]:
    del player_id
    action_logit = [
        (a, self._learning_rate * self.value(state, action=a) + np.log(p))
        for a, p in self._policy.action_probabilities(state).items()
    ]
    action, logit = zip(*action_logit)
    return dict(zip(action, mirror_descent.softmax_projection(logit)))


class MunchausenMirrorDescent(mirror_descent.MirrorDescent):
  """Munchausen Online Mirror Descent algorithm.

  This algorithm is equivalent to the online mirror descent algorithm but
  instead of summing value functions, it directly computes the cumulative
  Q-function using a penalty with respect to the previous policy.
  """

  def eval_state(self, state: pyspiel.State, learning_rate: float):
    """Evaluate the value of a state."""
    state_str = state.observation_string(pyspiel.PlayerId.DEFAULT_PLAYER_ID)
    # Return the already calculated value if present.
    if self._state_value.has(state_str):
      return self._state_value(state_str)
    # Otherwise, calculate the value of the state.
    v = self.get_state_value(state, learning_rate)
    self._state_value.set_value(state_str, v)
    return v

  def get_projected_policy(self) -> policy_lib.Policy:
    """Returns the projected policy."""
    return ProjectedPolicyMunchausen(self._game,
                                     list(range(self._game.num_players())),
                                     self._state_value, self._lr, self._policy)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""OpenSpiel API."""

import random
from typing import Any, List, Text, Tuple, Dict

from open_spiel.python.examples.meta_cfr.sequential_games import world_representation
import pyspiel


class WorldState(world_representation.WorldState):
  """World state representation for openspiel games.

  This class implements world_representation class for openspiel games.

  Attributes:
    game_name: Name of openspiel game we want to initialize.
    config: Config containing game parameters to initialize the game.
    state: Initial state of an openspeil game.
    chance_policy: The policy of the chance node in the game tree.
  """

  def __init__(self, game_name: str, config: Dict[str, Any],
               perturbation: bool, random_seed: int = 100):
    self._perturbation = perturbation
    self._history = []
    self._random_seed = random_seed
    self.game_name = game_name
    self.config = config
    self._game = pyspiel.load_game(self.game_name, self.config)
    if str(self._game.get_type().dynamics) == "Dynamics.SIMULTANEOUS":
      self._game = pyspiel.convert_to_turn_based(self._game)
    # initial_state
    self.state = self._game.new_initial_state()
    self.chance_policy = self.get_chance_policy()
    random.seed(self._random_seed)

  def get_distinct_actions(self) -> List[int]:
    """See base class."""
    return list(range(self._game.num_distinct_actions()))

  def is_terminal(self) -> bool:
    """See base class."""
    return self.state.is_terminal()

  def get_actions(self) -> List[Any]:
    """See base class."""
    if self.is_terminal():
      return [[], [], []]
    actions = [[0], [0], [0]]
    if self.state.is_chance_node():
      legal_actions = [
          action for (action, prob) in self.state.chance_outcomes()
      ]
    else:
      legal_actions = self.state.legal_actions()
    actions[self.state.current_player() + 1] = legal_actions
    return actions

  def get_infostate_string(self, player: int) -> Text:
    """See base class."""
    infostate = self.state.information_state_string(player - 1)
    return str(len(self._history)) + "|" + str(infostate)

  def apply_actions(self, actions: Tuple[int, int, int]) -> None:
    """See base class."""
    self.state.apply_action(actions[self.state.current_player() + 1])
    self.chance_policy = self.get_chance_policy()
    self._history.append(actions)

  def get_utility(self, player: int) -> float:
    """See base class."""
    assert self.is_terminal()
    return float(self.state.returns()[player - 1])

  def get_chance_policy(self) -> Dict[int, float]:
    """See base class."""
    if self.is_terminal():
      return {}

    if not self.state.is_chance_node():
      return {0: 1}

    chance_policy = {
        action: prob for (action, prob) in self.state.chance_outcomes()
    }

    if self._perturbation:
      probs = [random.random() for _ in self.state.chance_outcomes()]
      chance_policy = {
          action: probs[i] / sum(probs)
          for i, (action, prob) in enumerate(self.state.chance_outcomes())
      }

    return chance_policy



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A bot that asks the user which action to play."""

import math
import os

import pyspiel

_MAX_WIDTH = int(os.getenv("COLUMNS", "80"))  # Get your TTY width.


def _print_columns(strings):
  """Prints a list of strings in columns."""
  padding = 2
  longest = max(len(s) for s in strings)
  max_columns = math.floor((_MAX_WIDTH - 1) / (longest + 2 * padding))
  rows = math.ceil(len(strings) / max_columns)
  columns = math.ceil(len(strings) / rows)  # Might not fill all max_columns.
  for r in range(rows):
    for c in range(columns):
      i = r + c * rows
      if i < len(strings):
        print(" " * padding + strings[i].ljust(longest + padding), end="")
    print()


class HumanBot(pyspiel.Bot):
  """Asks the user which action to play."""

  def step_with_policy(self, state):
    """Returns the stochastic policy and selected action in the given state."""
    legal_actions = state.legal_actions(state.current_player())
    if not legal_actions:
      return [], pyspiel.INVALID_ACTION
    p = 1 / len(legal_actions)
    policy = [(action, p) for action in legal_actions]

    action_map = {
        state.action_to_string(state.current_player(), action): action
        for action in legal_actions
    }

    while True:
      action_str = input("Choose an action (empty to print legal actions): ")

      if not action_str:
        print("Legal actions(s):")
        longest_num = max(len(str(action)) for action in legal_actions)
        _print_columns([
            "{}: {}".format(str(action).rjust(longest_num), action_str)
            for action_str, action in sorted(action_map.items())
        ])
        continue

      if action_str in action_map:
        return policy, action_map[action_str]

      try:
        action = int(action_str)
      except ValueError:
        print("Could not parse the action:", action_str)
        continue

      if action in legal_actions:
        return policy, action

      print("Illegal action selected:", action_str)

  def step(self, state):
    return self.step_with_policy(state)[1]

  def restart_at(self, state):
    pass




# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple benchmark for MFG algorithms and environments."""

import itertools
import time
from typing import Sequence

from absl import app
from absl import flags

from open_spiel.python.mfg import games  # pylint: disable=unused-import
from open_spiel.python.mfg.algorithms import fictitious_play
import pyspiel

FLAGS = flags.FLAGS

flags.DEFINE_list('games',
                  ['python_mfg_crowd_modelling', 'mfg_crowd_modelling'],
                  'List of games to benchmark.')
flags.DEFINE_list(
    'parameters', ['size:10;100', 'horizon:10;100'],
    'List of parameters to sweep on (see default flag value for '
    'syntax).')


def convert_param_spec(param_spec):
  """Converts 'size:10;200' into ('size', [10, 200])."""
  split = param_spec.split(':', 2)
  return split[0], [int(v) for v in split[1].split(';')]


def main(argv: Sequence[str]) -> None:
  if len(argv) > 1:
    raise app.UsageError('Too many command-line arguments.')

  param_names, param_values = zip(
      *[convert_param_spec(spec) for spec in FLAGS.parameters])
  header = (['game_name'] + list(param_names) +
            ['fictitious_play_iteration_time'])
  timing_results = []
  for game_name in FLAGS.games:
    for param_tuple in itertools.product(*param_values):
      result_line = [game_name] + [str(p) for p in param_tuple]
      print('Computing timings for:', ' '.join(result_line))
      param_dict = dict(zip(param_names, param_tuple))
      game = pyspiel.load_game(game_name, param_dict)
      t0 = time.time()
      fp = fictitious_play.FictitiousPlay(game)
      fp.iteration()
      elapsed = time.time() - t0
      result_line.append(f'{elapsed:.4f}s')
      print(' '.join(result_line))
      timing_results.append(result_line)

  print('\nRESULTS:')
  print(' '.join(header))
  for line in timing_results:
    print(' '.join([str(v) for v in line]))


if __name__ == '__main__':
  app.run(main)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for fictitious play."""

from absl.testing import absltest
from absl.testing import parameterized

from open_spiel.python import policy
from open_spiel.python import rl_agent_policy
from open_spiel.python import rl_environment
from open_spiel.python.jax import dqn
from open_spiel.python.mfg import value
from open_spiel.python.mfg.algorithms import best_response_value
from open_spiel.python.mfg.algorithms import distribution
from open_spiel.python.mfg.algorithms import fictitious_play
from open_spiel.python.mfg.algorithms import greedy_policy
from open_spiel.python.mfg.algorithms import nash_conv
from open_spiel.python.mfg.algorithms import policy_value
from open_spiel.python.mfg.games import crowd_modelling
import pyspiel


class FictitiousPlayTest(parameterized.TestCase):

  @parameterized.named_parameters(("python", "python_mfg_crowd_modelling"),
                                  ("cpp", "mfg_crowd_modelling"))
  def test_run(self, name: str):
    """Checks if fictitious play works."""
    game = pyspiel.load_game(name)
    fp = fictitious_play.FictitiousPlay(game)
    for _ in range(10):
      fp.iteration()
    fp_policy = fp.get_policy()
    nash_conv_fp = nash_conv.NashConv(game, fp_policy)

    self.assertAlmostEqual(nash_conv_fp.nash_conv(), 0.991, places=3)

  @parameterized.named_parameters(("at_init", True), ("at_each_step", False))
  def test_learning_rate(self, at_init: bool):
    """Checks if learning rate works."""
    game = crowd_modelling.MFGCrowdModellingGame()
    lr = 1.0
    fp = fictitious_play.FictitiousPlay(game, lr=lr if at_init else None)
    for _ in range(10):
      fp.iteration(learning_rate=None if at_init else lr)
    fp_policy = fp.get_policy()
    nash_conv_fp = nash_conv.NashConv(game, fp_policy)

    self.assertAlmostEqual(nash_conv_fp.nash_conv(), 55.745, places=3)

  def test_soft_max(self):
    """Checks if soft-max policy works."""
    game = crowd_modelling.MFGCrowdModellingGame()
    fp = fictitious_play.FictitiousPlay(game, temperature=1)
    for _ in range(10):
      fp.iteration()
    fp_policy = fp.get_policy()
    nash_conv_fp = nash_conv.NashConv(game, fp_policy)

    self.assertAlmostEqual(nash_conv_fp.nash_conv(), 1.062, places=3)

  @parameterized.named_parameters(("python", "python_mfg_crowd_modelling"),
                                  ("cpp", "mfg_crowd_modelling"))
  def test_dqn(self, name):
    """Checks if fictitious play with DQN-based value function works."""
    game = pyspiel.load_game(name)
    dfp = fictitious_play.FictitiousPlay(game)

    uniform_policy = policy.UniformRandomPolicy(game)
    dist = distribution.DistributionPolicy(game, uniform_policy)
    envs = [
        rl_environment.Environment(
            game, mfg_distribution=dist, mfg_population=p)
        for p in range(game.num_players())
    ]
    dqn_agent = dqn.DQN(
        0,
        state_representation_size=envs[0].observation_spec()["info_state"][0],
        num_actions=envs[0].action_spec()["num_actions"],
        hidden_layers_sizes=[256, 128, 64],
        replay_buffer_capacity=100,
        batch_size=5,
        epsilon_start=0.02,
        epsilon_end=0.01)

    br_policy = rl_agent_policy.RLAgentPolicy(
        game, dqn_agent, 0, use_observation=True)
    for _ in range(10):
      dfp.iteration(br_policy=br_policy)

    dfp_policy = dfp.get_policy()
    nash_conv_dfp = nash_conv.NashConv(game, dfp_policy)

    self.assertAlmostEqual(nash_conv_dfp.nash_conv(), 1.056, places=3)

  def test_average(self):
    """Test the average of policies.

    Here we test that the average of values is the value of the average policy.
    """
    game = crowd_modelling.MFGCrowdModellingGame()
    uniform_policy = policy.UniformRandomPolicy(game)
    mfg_dist = distribution.DistributionPolicy(game, uniform_policy)
    br_value = best_response_value.BestResponse(
        game, mfg_dist, value.TabularValueFunction(game))
    py_value = policy_value.PolicyValue(game, mfg_dist, uniform_policy,
                                        value.TabularValueFunction(game))
    greedy_pi = greedy_policy.GreedyPolicy(game, None, br_value)
    greedy_pi = greedy_pi.to_tabular()
    merged_pi = fictitious_play.MergedPolicy(
        game, list(range(game.num_players())), [uniform_policy, greedy_pi],
        [mfg_dist, distribution.DistributionPolicy(game, greedy_pi)],
        [0.5, 0.5])
    merged_pi_value = policy_value.PolicyValue(game, mfg_dist, merged_pi,
                                               value.TabularValueFunction(game))

    self.assertAlmostEqual(
        merged_pi_value(game.new_initial_state()),
        (br_value(game.new_initial_state()) +
         py_value(game.new_initial_state())) / 2)


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Main file to train and evaluate meta-cfr agent, cfr and cfr-plus."""

from typing import Sequence

from absl import app
from absl import flags
import numpy as np

from open_spiel.python.examples.meta_cfr.sequential_games import cfr
from open_spiel.python.examples.meta_cfr.sequential_games import evaluation
from open_spiel.python.examples.meta_cfr.sequential_games import game_tree_utils
from open_spiel.python.examples.meta_cfr.sequential_games import meta_learning
from open_spiel.python.examples.meta_cfr.sequential_games import openspiel_api


FLAGS = flags.FLAGS

flags.DEFINE_integer("random_seed_size", 30, "Number of random seeds to use.")


def main(argv: Sequence[str]) -> None:
  del argv
  config = {"players": FLAGS.players}
  random_seeds_eval = np.random.choice(
      np.array(list(range(1000))), size=FLAGS.random_seed_size, replace=False)

  # Train a meta-cfr agent
  meta_cfr_agent = meta_learning.MetaCFRRegretAgent(
      training_epochs=1,
      meta_learner_training_epochs=FLAGS.meta_learner_training_epochs,
      game_name=FLAGS.game,
      game_config=config,
      perturbation=FLAGS.perturbation,
      seed=FLAGS.random_seed,
      model_type=FLAGS.model_type,
      best_response=True)
  meta_cfr_agent.train()

  cfr_vals = np.zeros((FLAGS.meta_learner_training_epochs,))
  cfr_plus_vals = np.zeros((FLAGS.meta_learner_training_epochs,))

  for seed in list(random_seeds_eval):

    # Evaluate a meta-cfr agent
    world_state = openspiel_api.WorldState(
        FLAGS.game, config, perturbation=True, random_seed=seed)
    meta_cfr_vals = evaluation.CFRBREvaluation(meta_cfr_agent, world_state)

    # Evaluate a cfr plus agent
    game_tree = game_tree_utils.build_game_tree(
        openspiel_api.WorldState(
            FLAGS.game,
            config,
            perturbation=FLAGS.perturbation,
            random_seed=seed))
    _, cfr_plus_vals = cfr.compute_cfr_plus_values(
        game_tree, FLAGS.meta_learner_training_epochs)

    # Evaluate a cfr agent
    game_tree = game_tree_utils.build_game_tree(
        openspiel_api.WorldState(
            FLAGS.game,
            config,
            perturbation=FLAGS.perturbation,
            random_seed=seed))
    _, cfr_vals = cfr.compute_cfr_values(
        game_tree, FLAGS.meta_learner_training_epochs)

  print("Evaluation seed:", random_seeds_eval)
  print("Meta_cfr agent:", meta_cfr_vals)
  print("cfr_plus agent:", cfr_plus_vals)
  print("cfr agent:", cfr_vals)


if __name__ == "__main__":
  app.run(main)



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""An RL agent wrapper for the MCTS bot."""

import numpy as np
from open_spiel.python import rl_agent
import pyspiel


class MCTSAgent(rl_agent.AbstractAgent):
  """MCTS agent class.

  Important note: this agent requires the environment to provide the full state
  in its TimeStep objects. Hence, the environment must be created with the
  use_full_state flag set to True, and the state must be serializable.
  """

  def __init__(self, player_id, num_actions, mcts_bot, name="mcts_agent"):
    assert num_actions > 0
    self._player_id = player_id
    self._mcts_bot = mcts_bot
    self._num_actions = num_actions

  def step(self, time_step, is_evaluation=False):
    # If it is the end of the episode, don't select an action.
    if time_step.last():
      return

    assert "serialized_state" in time_step.observations
    _, state = pyspiel.deserialize_game_and_state(
        time_step.observations["serialized_state"])

    # Call the MCTS bot's step to get the action.
    probs = np.zeros(self._num_actions)
    action = self._mcts_bot.step(state)
    probs[action] = 1.0

    return rl_agent.StepOutput(action=action, probs=probs)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Liar's Poker."""

import pickle

from absl.testing import absltest
import numpy as np

from open_spiel.python.games import liars_poker
import pyspiel


class LiarsPokerTest(absltest.TestCase):

  def test_can_create_game_and_state(self):
    """Checks we can create the game and a state."""
    game = liars_poker.LiarsPoker({"hand_length": 3, "num_digits": 3})
    state = game.new_initial_state()
    # Ensure no moves have been made.
    expected_hands = [[] for _ in range(game.num_players())]
    expected_bidder = -1
    expected_current_player = pyspiel.PlayerId.CHANCE
    expected_current_count = "None"
    expected_current_number = "None"
    expected_rebid = False
    expected = (
        "Hands: {}, Bidder: {}, Current Player: {}, Current Bid: {} of {},"
        " Rebid: {}".format(
            expected_hands,
            expected_bidder,
            expected_current_player,
            expected_current_count,
            expected_current_number,
            expected_rebid,
        )
    )
    self.assertEqual(str(state), expected)

  def test_draw_hands(self):
    """Tests hand drawing functions."""
    game = liars_poker.LiarsPoker({"hand_length": 3, "num_digits": 3})
    state = game.new_initial_state()
    expected_hands = [[] for _ in range(game.num_players())]
    for i in range(game.num_players() * game.hand_length):
      # Verify we have chance nodes until all player hands are filled.
      self.assertEqual(state.current_player(), pyspiel.PlayerId.CHANCE)

      # Draw a digit.
      outcomes_with_probs = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes_with_probs)
      action = np.random.choice(action_list, p=prob_list)

      # Verify players' hands are filled correctly.
      cur_player = i % game.num_players()
      expected_hands[cur_player].append(action)
      state.apply_action(action)
      self.assertEqual(state.hands, expected_hands)
    # Assert after all hands are filled, we have non-chance nodes.
    cur_player = state.current_player()
    self.assertNotEqual(cur_player, pyspiel.PlayerId.CHANCE)
    self.assertEqual(cur_player, 0)

  def _populate_game_hands(self, game, state):
    """Populates players hands for testing."""
    for _ in range(game.num_players() * game.hand_length):
      outcomes_with_probs = state.chance_outcomes()
      action_list, prob_list = zip(*outcomes_with_probs)
      action = np.random.choice(action_list, p=prob_list)
      state.apply_action(action)

  def test_basic_bid(self):
    """Tests a single bid."""
    game = liars_poker.LiarsPoker({"hand_length": 3, "num_digits": 3})
    state = game.new_initial_state()
    expected_bid_history = np.zeros(
        (state.total_possible_bids, state.num_players())
    )

    # Fill players hands.
    self._populate_game_hands(game, state)
    # After all hands are filled, have player 0 bid.
    cur_player = state.current_player()
    action = 2
    state.apply_action(action)

    # Verify bid history is updated correctly.
    bid_offset = liars_poker.BID_ACTION_OFFSET
    expected_bid_history[action - bid_offset][cur_player] = 1
    self.assertTrue((state.bid_history == expected_bid_history).all())

    # Verify next set of legal bids is greater than the current bid.
    for next_action in state.legal_actions():
      if next_action == liars_poker.CHALLENGE_ACTION:
        continue
      self.assertGreater(next_action, action)

  def _verify_returns(self, game, state):
    self.assertTrue(state.winner() != -1 or state.loser() != -1)
    actual_returns = state.returns()
    if state.winner() != -1:
      expected_returns = [-1.0 for _ in range(game.num_players())]
      expected_returns[state.winner()] = game.num_players() - 1
    else:
      expected_returns = [1.0 for _ in range(game.num_players())]
      expected_returns[state.loser()] = -1.0 * (game.num_players() - 1)
    self.assertEqual(actual_returns, expected_returns)

  def test_single_random_round(self):
    """Runs a single round of bidding followed by a challenge."""
    game = liars_poker.LiarsPoker({"hand_length": 3, "num_digits": 3})
    state = game.new_initial_state()
    expected_challenge_history = np.zeros(
        (state.total_possible_bids, state.num_players())
    )

    # Fill players hands.
    self._populate_game_hands(game, state)
    # Have player 0 bid.
    action = 2
    state.apply_action(action)
    # Verify challenge action is available to the next player.
    challenge = liars_poker.CHALLENGE_ACTION
    self.assertIn(challenge, state.legal_actions())
    # Player 1 challenges.
    cur_player = state.current_player()
    state.apply_action(challenge)
    bid_offset = liars_poker.BID_ACTION_OFFSET
    expected_challenge_history[action - bid_offset][cur_player] = 1
    # Verify challenge history is updated correctly.
    self.assertTrue(
        (state.challenge_history == expected_challenge_history).all()
    )
    # Original bidder challenges, thus agreeing to a count.
    cur_player = state.current_player()
    state.apply_action(challenge)
    expected_challenge_history[action - bid_offset][cur_player] = 1
    # Verify challenge history is updated correctly.
    self.assertTrue(
        (state.challenge_history == expected_challenge_history).all()
    )

    # Verify game is over.
    self.assertTrue(state.is_terminal())
    # Verify returns.
    self._verify_returns(game, state)

  def test_single_deterministic_round(self):
    """Runs a single round where cards are dealt deterministically."""
    game = liars_poker.LiarsPoker({"hand_length": 3, "num_digits": 3})
    state = game.new_initial_state()

    # Deal player 0 all "1" cards and player 1 all "2" cards.
    for i in range(game.num_players() * game.hand_length):
      if i % 2 == 0:
        # Deal card to player 0
        state.apply_action(1)
      else:
        # Deal card to player 1
        state._apply_action(2)

    # Have player 0 bid that there are four 1's.
    state.apply_action(state.encode_bid(4, 1) + liars_poker.BID_ACTION_OFFSET)
    # Player 1 challenges.
    state.apply_action(liars_poker.CHALLENGE_ACTION)
    # Player 0 accepts the challenge.
    state.apply_action(liars_poker.CHALLENGE_ACTION)
    # Verify game ends with player 0 losing.
    self.assertTrue(state.is_terminal())
    self.assertEqual(state.loser(), 0)
    expected_returns = [1.0 for _ in range(game.num_players())]
    expected_returns[state.loser()] = -1.0 * (game.num_players() - 1)
    self.assertEqual(state.returns(), expected_returns)

  def test_single_rebid(self):
    """Runs a 2 player game where a rebid is enacted."""
    game = liars_poker.LiarsPoker({"hand_length": 3, "num_digits": 3})
    state = game.new_initial_state()

    # Fill players hands.
    self._populate_game_hands(game, state)
    # Have player 0 bid.
    state.apply_action(2)
    # Player 1 challenges.
    state.apply_action(liars_poker.CHALLENGE_ACTION)
    # Original bidder rebids.
    state.apply_action(3)
    # Verify game is not over.
    self.assertFalse(state.is_terminal())
    self.assertEqual(state.returns(), [0.0 for _ in range(game.num_players())])
    # Player 1 challenges again.
    state.apply_action(liars_poker.CHALLENGE_ACTION)

    # Verify game is now over.
    self.assertTrue(state.is_terminal())
    self._verify_returns(game, state)

  def test_rebid_then_new_bid(self):
    """Runs a 2 player game where a rebid is enacted."""
    game = liars_poker.LiarsPoker({"hand_length": 3, "num_digits": 3})
    state = game.new_initial_state()

    # Fill players hands.
    self._populate_game_hands(game, state)
    # Have player 0 bid.
    state.apply_action(2)
    # Player 1 challenges.
    state.apply_action(liars_poker.CHALLENGE_ACTION)
    # Original bidder rebids.
    state.apply_action(3)
    # Verify game is not over.
    self.assertFalse(state.is_terminal())
    self.assertEqual(state.returns(), [0.0 for _ in range(game.num_players())])
    # Player 1 bids.
    state.apply_action(4)
    # Verify game is not over.
    self.assertFalse(state.is_terminal())
    # Player 0 challenges.
    state.apply_action(liars_poker.CHALLENGE_ACTION)
    # Verify we're not rebidding and counts is only called once both players
    # challenge.
    self.assertFalse(state.is_terminal())
    # Player 1 challenges and ends the game with a counts.
    state.apply_action(liars_poker.CHALLENGE_ACTION)

    # Verify game is now over.
    self.assertTrue(state.is_terminal())
    self._verify_returns(game, state)

  def test_game_from_cc(self):
    """Runs the standard game tests, checking API consistency."""
    game = pyspiel.load_game("python_liars_poker", {"players": 2})
    pyspiel.random_sim_test(game, num_sims=10, serialize=False, verbose=True)

  def test_pickle(self):
    """Checks pickling and unpickling of game and state."""
    game = pyspiel.load_game("python_liars_poker")
    pickled_game = pickle.dumps(game)
    unpickled_game = pickle.loads(pickled_game)
    self.assertEqual(str(game), str(unpickled_game))
    state = game.new_initial_state()
    for a in [2, 3, 4, 5]:
      state.apply_action(a)
    ser_str = pyspiel.serialize_game_and_state(game, state)
    new_game, new_state = pyspiel.deserialize_game_and_state(ser_str)
    self.assertEqual(str(game), str(new_game))
    self.assertEqual(str(state), str(new_state))
    pickled_state = pickle.dumps(state)
    unpickled_state = pickle.loads(pickled_state)
    self.assertEqual(str(state), str(unpickled_state))

  def test_cloned_state_matches_original_state(self):
    """Check we can clone states successfully."""
    game = liars_poker.LiarsPoker({"hand_length": 3, "num_digits": 3})
    state = game.new_initial_state()
    state.apply_action(1)
    state.apply_action(2)
    clone = state.clone()

    self.assertEqual(state.history(), clone.history())
    self.assertEqual(state.num_players(), clone.num_players())
    self.assertEqual(state.move_number(), clone.move_number())
    self.assertEqual(state.num_distinct_actions(), clone.num_distinct_actions())

    self.assertEqual(state._current_player, clone._current_player)
    self.assertEqual(state._current_action, clone._current_action)
    np.testing.assert_array_equal(state.bid_history, clone.bid_history)
    np.testing.assert_array_equal(
        state.challenge_history, clone.challenge_history
    )


if __name__ == "__main__":
  absltest.main()



# Copyright 2022 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Useful sequence form functions used in the MMD implementation."""

import numpy as np
from open_spiel.python import policy


_DELIMITER = " -=- "
_EMPTY_INFOSET_KEYS = ["***EMPTY_INFOSET_P0***", "***EMPTY_INFOSET_P1***"]
_EMPTY_INFOSET_ACTION_KEYS = [
    "***EMPTY_INFOSET_ACTION_P0***", "***EMPTY_INFOSET_ACTION_P1***"
]


def _get_isa_key(info_state, action):
  return info_state + _DELIMITER + str(action)


def _get_action_from_key(isa_key):
  _, action_str = isa_key.split(_DELIMITER)
  return int(action_str)


def _get_infostate_from_key(isa_key):
  assert not is_root(isa_key), "Cannot use this method for root nodes."
  infostate, _ = isa_key.split(_DELIMITER)
  return infostate


def is_root(key):
  return True if key in _EMPTY_INFOSET_KEYS + _EMPTY_INFOSET_ACTION_KEYS else False


def construct_vars(game):
  """Construct useful sequence from variables from game.

  Args:
      game: The spiel game to solve (must be zero-sum, sequential, and have
        chance node of deterministic or explicit stochastic).

  Returns:
      An 8 tuple of sequence form variables from _construct_vars by
      recursively
      traversing the game tree.

  """

  initial_state = game.new_initial_state()

  # initialize variables
  infosets = [{_EMPTY_INFOSET_KEYS[0]: 0}, {_EMPTY_INFOSET_KEYS[1]: 0}]
  infoset_actions_to_seq = [{
      _EMPTY_INFOSET_ACTION_KEYS[0]: 0
  }, {
      _EMPTY_INFOSET_ACTION_KEYS[1]: 0
  }]
  infoset_action_maps = [{
      _EMPTY_INFOSET_KEYS[0]: [_EMPTY_INFOSET_ACTION_KEYS[0]]
  }, {
      _EMPTY_INFOSET_KEYS[1]: [_EMPTY_INFOSET_ACTION_KEYS[1]]
  }]

  # infoset_action_maps = [{}, {}]
  payoff_dict = dict()

  infoset_parent_map = [{
      _EMPTY_INFOSET_ACTION_KEYS[0]: None
  }, {
      _EMPTY_INFOSET_ACTION_KEYS[1]: None
  }]
  infoset_actions_children = [{
      _EMPTY_INFOSET_ACTION_KEYS[0]: []
  }, {
      _EMPTY_INFOSET_ACTION_KEYS[1]: []
  }]

  _construct_vars(initial_state, infosets, infoset_actions_to_seq,
                  infoset_action_maps, infoset_parent_map, 1.0,
                  _EMPTY_INFOSET_KEYS[:], _EMPTY_INFOSET_ACTION_KEYS[:],
                  payoff_dict, infoset_actions_children)

  payoff_mat = _construct_numpy_vars(payoff_dict, infoset_actions_to_seq)
  return (infosets, infoset_actions_to_seq,
          infoset_action_maps, infoset_parent_map,
          payoff_mat, infoset_actions_children)


def uniform_random_seq(game, infoset_actions_to_seq):
  """Generate uniform random sequence.

  The sequence generated is equivalent to a uniform random tabular policy.

  Args:
      game: the spiel game to solve (must be zero-sum, sequential, and have
        chance mode of deterministic or explicit stochastic).
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.

  Returns:
      A list of NumPy arrays, one for each player.
  """
  policies = policy.TabularPolicy(game)
  initial_state = game.new_initial_state()
  sequences = [
      np.ones(len(infoset_actions_to_seq[0])),
      np.ones(len(infoset_actions_to_seq[1]))
  ]
  _policy_to_sequence(initial_state, policies, sequences,
                      infoset_actions_to_seq, [1, 1])
  return sequences


def _construct_vars(state, infosets, infoset_actions_to_seq,
                    infoset_action_maps, infoset_parent_map, chance_reach,
                    parent_is_keys, parent_isa_keys, payoff_dict,
                    infoset_actions_children):
  """Recursively builds maps and the sequence form payoff matrix.

  Args:
      state: pyspiel (OpenSpiel) state
      infosets: a list of dicts, one per player, that maps infostate to an id.
        The dicts are filled by this function and should initially only
        contain root values.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id. The dicts are filled by
        this function and should inirially only contain the root values.
      infoset_action_maps: a list of dicts, one per player, that maps each
        info_state to a list of (infostate, action) string.
      infoset_parent_map: a list of dicts, one per player, that maps each
        info_state to an (infostate, action) string.
      chance_reach: the contribution of chance's reach probability (should
        start at 1).
      parent_is_keys: a list of parent information state keys for this state
      parent_isa_keys: a list of parent (infostate, action) keys
      payoff_dict: a dict that maps ((infostate, action), (infostate, action))
        to the chance weighted reward
      infoset_actions_children: a list of dicts, one for each player, mapping
        (infostate, action) keys to reachable infostates for each player
  """

  if state.is_terminal():
    returns = state.returns()
    matrix_index = (parent_isa_keys[0], parent_isa_keys[1])
    payoff_dict.setdefault(matrix_index, 0)
    # note the payoff matrix A is for the min max problem x.T @ A y
    # where x is player 0 in openspiel
    payoff_dict[matrix_index] += -returns[0] * chance_reach
    return

  if state.is_chance_node():
    for action, prob in state.chance_outcomes():
      new_state = state.child(action)
      _construct_vars(new_state, infosets, infoset_actions_to_seq,
                      infoset_action_maps, infoset_parent_map,
                      prob * chance_reach, parent_is_keys, parent_isa_keys,
                      payoff_dict, infoset_actions_children)
    return

  player = state.current_player()
  info_state = state.information_state_string(player)
  legal_actions = state.legal_actions(player)

  # Add to the infostate maps
  if info_state not in infosets[player]:
    infosets[player][info_state] = len(infosets[player])
  if info_state not in infoset_action_maps[player]:
    infoset_action_maps[player][info_state] = []

  # Add to infoset to parent infoset action map
  if info_state not in infoset_parent_map[player]:
    infoset_parent_map[player][info_state] = parent_isa_keys[player]

  # add as child to parent
  if parent_isa_keys[player] in infoset_actions_children[player]:
    if info_state not in infoset_actions_children[player][
        parent_isa_keys[player]]:
      infoset_actions_children[player][parent_isa_keys[player]].append(
          info_state)
  else:
    infoset_actions_children[player][parent_isa_keys[player]] = [info_state]

  new_parent_is_keys = parent_is_keys[:]
  new_parent_is_keys[player] = info_state

  for action in legal_actions:
    isa_key = _get_isa_key(info_state, action)
    if isa_key not in infoset_actions_to_seq[player]:
      infoset_actions_to_seq[player][isa_key] = len(
          infoset_actions_to_seq[player])
    if isa_key not in infoset_action_maps[player][info_state]:
      infoset_action_maps[player][info_state].append(isa_key)

    new_parent_isa_keys = parent_isa_keys[:]
    new_parent_isa_keys[player] = isa_key
    new_state = state.child(action)
    _construct_vars(new_state, infosets, infoset_actions_to_seq,
                    infoset_action_maps, infoset_parent_map, chance_reach,
                    new_parent_is_keys, new_parent_isa_keys, payoff_dict,
                    infoset_actions_children)


def _construct_numpy_vars(payoff_dict, infoset_actions_to_seq):
  """Convert sequence form payoff dict to numpy array.

  Args:
      payoff_dict: a dict that maps ((infostate, action), (infostate, action))
        to the chance weighted reward.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.

  Returns:
      A numpy array corresponding to the chance weighted rewards
      i.e. the sequence form payoff matrix.

  """
  sequence_sizes = (len(infoset_actions_to_seq[0]),
                    len(infoset_actions_to_seq[1]))
  payoff_mat = np.zeros(sequence_sizes)
  for p1_sequence, i in infoset_actions_to_seq[0].items():
    for p2_sequence, j in infoset_actions_to_seq[1].items():
      payoff_mat[i, j] = payoff_dict.get((p1_sequence, p2_sequence), 0)
  return payoff_mat


def sequence_to_policy(sequences, game, infoset_actions_to_seq,
                       infoset_action_maps):
  """Convert sequence form policies to the realization-equivalent tabular ones.

  Args:
      sequences: list of two sequence form policies, one for each player.
      game: a spiel game with two players.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.
      infoset_action_maps: a list of dicts, one per player, that maps each
        info_state to a list of (infostate, action) string.

  Returns:
      A TabularPolicy object.
  """

  policies = policy.TabularPolicy(game)
  for player in range(2):
    for info_state in infoset_action_maps[player]:
      if is_root(info_state):
        continue

      state_policy = policies.policy_for_key(info_state)
      total_weight = 0
      num_actions = 0

      for isa_key in infoset_action_maps[player][info_state]:
        total_weight += sequences[player][infoset_actions_to_seq[player]
                                          [isa_key]]
        num_actions += 1

      unif_pr = 1.0 / num_actions
      for isa_key in infoset_action_maps[player][info_state]:
        rel_weight = sequences[player][infoset_actions_to_seq[player][isa_key]]
        _, action_str = isa_key.split(_DELIMITER)
        action = int(action_str)
        pr_action = rel_weight / total_weight if total_weight > 0 else unif_pr
        state_policy[action] = pr_action
  return policies


def policy_to_sequence(game, policies, infoset_actions_to_seq):
  """Converts a TabularPolicy object for a two-player game.

  The converted policy is its realization-equivalent sequence form one.

  Args:
      game: a two-player open spiel game.
      policies: a TabularPolicy object.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.

  Returns:
      A list of numpy arrays, one for each player.
  """
  initial_state = game.new_initial_state()
  sequences = [
      np.ones(len(infoset_actions_to_seq[0])),
      np.ones(len(infoset_actions_to_seq[1]))
  ]
  _policy_to_sequence(initial_state, policies, sequences,
                      infoset_actions_to_seq, [1, 1])
  return sequences


def _policy_to_sequence(state, policies, sequences, infoset_actions_to_seq,
                        parent_seq_val):
  """Converts a TabularPolicy object to its equivalent sequence form.

  This method modifies the sequences inplace and should not be called directly.

  Args:
      state: an openspiel state.
      policies: a TabularPolicy object.
      sequences: list of numpy arrays to be modified.
      infoset_actions_to_seq: a list of dicts, one per player, that maps a
        string of (infostate, action) pair to an id.
      parent_seq_val: list of parent sequence values, this method should be
        called with initial value of [1,1].
  """

  if state.is_terminal():
    return

  if state.is_chance_node():
    for action, _ in state.chance_outcomes():
      new_state = state.child(action)
      _policy_to_sequence(new_state, policies, sequences,
                          infoset_actions_to_seq, parent_seq_val)
    return

  player = state.current_player()
  info_state = state.information_state_string(player)
  legal_actions = state.legal_actions(player)
  state_policy = policies.policy_for_key(info_state)
  for action in legal_actions:
    isa_key = _get_isa_key(info_state, action)
    # update sequence form
    sequences[player][infoset_actions_to_seq[player]
                      [isa_key]] = parent_seq_val[player] * state_policy[action]
    new_parent_seq_val = parent_seq_val[:]
    new_parent_seq_val[player] = sequences[player][
        infoset_actions_to_seq[player][isa_key]]
    new_state = state.child(action)
    _policy_to_sequence(new_state, policies, sequences, infoset_actions_to_seq,
                        new_parent_seq_val)



# Copyright 2019 DeepMind Technologies Limited
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Monte-Carlo Tree Search algorithm for game play."""

import math
import time

import numpy as np

import pyspiel


class Evaluator(object):
  """Abstract class representing an evaluation function for a game.

  The evaluation function takes in an intermediate state in the game and returns
  an evaluation of that state, which should correlate with chances of winning
  the game. It returns the evaluation from all player's perspectives.
  """

  def evaluate(self, state):
    """Returns evaluation on given state."""
    raise NotImplementedError

  def prior(self, state):
    """Returns a probability for each legal action in the given state."""
    raise NotImplementedError


class RandomRolloutEvaluator(Evaluator):
  """A simple evaluator doing random rollouts.

  This evaluator returns the average outcome of playing random actions from the
  given state until the end of the game.  n_rollouts is the number of random
  outcomes to be considered.
  """

  def __init__(self, n_rollouts=1, random_state=None):
    self.n_rollouts = n_rollouts
    self._random_state = random_state or np.random.RandomState()

  def evaluate(self, state):
    """Returns evaluation on given state."""
    result = None
    for _ in range(self.n_rollouts):
      working_state = state.clone()
      while not working_state.is_terminal():
        if working_state.is_chance_node():
          outcomes = working_state.chance_outcomes()
          action_list, prob_list = zip(*outcomes)
          action = self._random_state.choice(action_list, p=prob_list)
        else:
          action = self._random_state.choice(working_state.legal_actions())
        working_state.apply_action(action)
      returns = np.array(working_state.returns())
      result = returns if result is None else result + returns

    return result / self.n_rollouts

  def prior(self, state):
    """Returns equal probability for all actions."""
    if state.is_chance_node():
      return state.chance_outcomes()
    else:
      legal_actions = state.legal_actions(state.current_player())
      return [(action, 1.0 / len(legal_actions)) for action in legal_actions]


class SearchNode(object):
  """A node in the search tree.

  A SearchNode represents a state and possible continuations from it. Each child
  represents a possible action, and the expected result from doing so.

  Attributes:
    action: The action from the parent node's perspective. Not important for the
      root node, as the actions that lead to it are in the past.
    player: Which player made this action.
    prior: A prior probability for how likely this action will be selected.
    explore_count: How many times this node was explored.
    total_reward: The sum of rewards of rollouts through this node, from the
      parent node's perspective. The average reward of this node is
      `total_reward / explore_count`
    outcome: The rewards for all players if this is a terminal node or the
      subtree has been proven, otherwise None.
    children: A list of SearchNodes representing the possible actions from this
      node, along with their expected rewards.
  """
  __slots__ = [
      "action",
      "player",
      "prior",
      "explore_count",
      "total_reward",
      "outcome",
      "children",
  ]

  def __init__(self, action, player, prior):
    self.action = action
    self.player = player
    self.prior = prior
    self.explore_count = 0
    self.total_reward = 0.0
    self.outcome = None
    self.children = []

  def uct_value(self, parent_explore_count, uct_c):
    """Returns the UCT value of child."""
    if self.outcome is not None:
      return self.outcome[self.player]

    if self.explore_count == 0:
      return float("inf")

    return self.total_reward / self.explore_count + uct_c * math.sqrt(
        math.log(parent_explore_count) / self.explore_count)

  def puct_value(self, parent_explore_count, uct_c):
    """Returns the PUCT value of child."""
    if self.outcome is not None:
      return self.outcome[self.player]

    return ((self.explore_count and self.total_reward / self.explore_count) +
            uct_c * self.prior * math.sqrt(parent_explore_count) /
            (self.explore_count + 1))

  def sort_key(self):
    """Returns the best action from this node, either proven or most visited.

    This ordering leads to choosing:
    - Highest proven score > 0 over anything else, including a promising but
      unproven action.
    - A proven draw only if it has higher exploration than others that are
      uncertain, or the others are losses.
    - Uncertain action with most exploration over loss of any difficulty
    - Hardest loss if everything is a loss
    - Highest expected reward if explore counts are equal (unlikely).
    - Longest win, if multiple are proven (unlikely due to early stopping).
    """
    return (0 if self.outcome is None else self.outcome[self.player],
            self.explore_count, self.total_reward)

  def best_child(self):
    """Returns the best child in order of the sort key."""
    return max(self.children, key=SearchNode.sort_key)

  def children_str(self, state=None):
    """Returns the string representation of this node's children.

    They are ordered based on the sort key, so order of being chosen to play.

    Args:
      state: A `pyspiel.State` object, to be used to convert the action id into
        a human readable format. If None, the action integer id is used.
    """
    return "\n".join([
        c.to_str(state)
        for c in reversed(sorted(self.children, key=SearchNode.sort_key))
    ])

  def to_str(self, state=None):
    """Returns the string representation of this node.

    Args:
      state: A `pyspiel.State` object, to be used to convert the action id into
        a human readable format. If None, the action integer id is used.
    """
    action = (
        state.action_to_string(state.current_player(), self.action)
        if state and self.action is not None else str(self.action))
    return ("{:>6}: player: {}, prior: {:5.3f}, value: {:6.3f}, sims: {:5d}, "
            "outcome: {}, {:3d} children").format(
                action, self.player, self.prior, self.explore_count and
                self.total_reward / self.explore_count, self.explore_count,
                ("{:4.1f}".format(self.outcome[self.player])
                 if self.outcome else "none"), len(self.children))

  def __str__(self):
    return self.to_str(None)


class MCTSBot(pyspiel.Bot):
  """Bot that uses Monte-Carlo Tree Search algorithm."""

  def __init__(self,
               game,
               uct_c,
               max_simulations,
               evaluator,
               solve=True,
               random_state=None,
               child_selection_fn=SearchNode.uct_value,
               dirichlet_noise=None,
               verbose=False,
               dont_return_chance_node=False):
    """Initializes a MCTS Search algorithm in the form of a bot.

    In multiplayer games, or non-zero-sum games, the players will play the
    greedy strategy.

    Args:
      game: A pyspiel.Game to play.
      uct_c: The exploration constant for UCT.
      max_simulations: How many iterations of MCTS to perform. Each simulation
        will result in one call to the evaluator. Memory usage should grow
        linearly with simulations * branching factor. How many nodes in the
        search tree should be evaluated. This is correlated with memory size and
        tree depth.
      evaluator: A `Evaluator` object to use to evaluate a leaf node.
      solve: Whether to back up solved states.
      random_state: An optional numpy RandomState to make it deterministic.
      child_selection_fn: A function to select the child in the descent phase.
        The default is UCT.
      dirichlet_noise: A tuple of (epsilon, alpha) for adding dirichlet noise to
        the policy at the root. This is from the alpha-zero paper.
      verbose: Whether to print information about the search tree before
        returning the action. Useful for confirming the search is working
        sensibly.
      dont_return_chance_node: If true, do not stop expanding at chance nodes.
        Enabled for AlphaZero.

    Raises:
      ValueError: if the game type isn't supported.
    """
    pyspiel.Bot.__init__(self)
    # Check that the game satisfies the conditions for this MCTS implemention.
    game_type = game.get_type()
    if game_type.reward_model != pyspiel.GameType.RewardModel.TERMINAL:
      raise ValueError("Game must have terminal rewards.")
    if game_type.dynamics != pyspiel.GameType.Dynamics.SEQUENTIAL:
      raise ValueError("Game must have sequential turns.")

    self._game = game
    self.uct_c = uct_c
    self.max_simulations = max_simulations
    self.evaluator = evaluator
    self.verbose = verbose
    self.solve = solve
    self.max_utility = game.max_utility()
    self._dirichlet_noise = dirichlet_noise
    self._random_state = random_state or np.random.RandomState()
    self._child_selection_fn = child_selection_fn
    self.dont_return_chance_node = dont_return_chance_node

  def restart_at(self, state):
    pass

  def step_with_policy(self, state):
    """Returns bot's policy and action at given state."""
    t1 = time.time()
    root = self.mcts_search(state)

    best = root.best_child()

    if self.verbose:
      seconds = time.time() - t1
      print("Finished {} sims in {:.3f} secs, {:.1f} sims/s".format(
          root.explore_count, seconds, root.explore_count / seconds))
      print("Root:")
      print(root.to_str(state))
      print("Children:")
      print(root.children_str(state))
      if best.children:
        chosen_state = state.clone()
        chosen_state.apply_action(best.action)
        print("Children of chosen:")
        print(best.children_str(chosen_state))

    mcts_action = best.action

    policy = [(action, (1.0 if action == mcts_action else 0.0))
              for action in state.legal_actions(state.current_player())]

    return policy, mcts_action

  def step(self, state):
    return self.step_with_policy(state)[1]

  def _apply_tree_policy(self, root, state):
    """Applies the UCT policy to play the game until reaching a leaf node.

    A leaf node is defined as a node that is terminal or has not been evaluated
    yet. If it reaches a node that has been evaluated before but hasn't been
    expanded, then expand it's children and continue.

    Args:
      root: The root node in the search tree.
      state: The state of the game at the root node.

    Returns:
      visit_path: A list of nodes descending from the root node to a leaf node.
      working_state: The state of the game at the leaf node.
    """
    visit_path = [root]
    working_state = state.clone()
    current_node = root
    while (not working_state.is_terminal() and
           current_node.explore_count > 0) or (
               working_state.is_chance_node() and self.dont_return_chance_node):
      if not current_node.children:
        # For a new node, initialize its state, then choose a child as normal.
        legal_actions = self.evaluator.prior(working_state)
        if current_node is root and self._dirichlet_noise:
          epsilon, alpha = self._dirichlet_noise
          noise = self._random_state.dirichlet([alpha] * len(legal_actions))
          legal_actions = [(a, (1 - epsilon) * p + epsilon * n)
                           for (a, p), n in zip(legal_actions, noise)]
        # Reduce bias from move generation order.
        self._random_state.shuffle(legal_actions)
        player = working_state.current_player()
        current_node.children = [
            SearchNode(action, player, prior) for action, prior in legal_actions
        ]

      if working_state.is_chance_node():
        # For chance nodes, rollout according to chance node's probability
        # distribution
        outcomes = working_state.chance_outcomes()
        action_list, prob_list = zip(*outcomes)
        action = self._random_state.choice(action_list, p=prob_list)
        chosen_child = next(
            c for c in current_node.children if c.action == action)
      else:
        # Otherwise choose node with largest UCT value
        chosen_child = max(
            current_node.children,
            key=lambda c: self._child_selection_fn(  # pylint: disable=g-long-lambda
                c, current_node.explore_count, self.uct_c))

      working_state.apply_action(chosen_child.action)
      current_node = chosen_child
      visit_path.append(current_node)

    return visit_path, working_state

  def mcts_search(self, state):
    """A vanilla Monte-Carlo Tree Search algorithm.

    This algorithm searches the game tree from the given state.
    At the leaf, the evaluator is called if the game state is not terminal.
    A total of max_simulations states are explored.

    At every node, the algorithm chooses the action with the highest PUCT value,
    defined as: `Q/N + c * prior * sqrt(parent_N) / N`, where Q is the total
    reward after the action, and N is the number of times the action was
    explored in this position. The input parameter c controls the balance
    between exploration and exploitation; higher values of c encourage
    exploration of under-explored nodes. Unseen actions are always explored
    first.

    At the end of the search, the chosen action is the action that has been
    explored most often. This is the action that is returned.

    This implementation supports sequential n-player games, with or without
    chance nodes. All players maximize their own reward and ignore the other
    players' rewards. This corresponds to max^n for n-player games. It is the
    norm for zero-sum games, but doesn't have any special handling for
    non-zero-sum games. It doesn't have any special handling for imperfect
    information games.

    The implementation also supports backing up solved states, i.e. MCTS-Solver.
    The implementation is general in that it is based on a max^n backup (each
    player greedily chooses their maximum among proven children values, or there
    exists one child whose proven value is game.max_utility()), so it will work
    for multiplayer, general-sum, and arbitrary payoff games (not just win/loss/
    draw games). Also chance nodes are considered proven only if all children
    have the same value.

    Some references:
    - Sturtevant, An Analysis of UCT in Multi-Player Games,  2008,
      https://web.cs.du.edu/~sturtevant/papers/multi-player_UCT.pdf
    - Nijssen, Monte-Carlo Tree Search for Multi-Player Games, 2013,
      https://project.dke.maastrichtuniversity.nl/games/files/phd/Nijssen_thesis.pdf
    - Silver, AlphaGo Zero: Starting from scratch, 2017
      https://deepmind.com/blog/article/alphago-zero-starting-scratch
    - Winands, Bjornsson, and Saito, "Monte-Carlo Tree Search Solver", 2008.
      https://dke.maastrichtuniversity.nl/m.winands/documents/uctloa.pdf

    Arguments:
      state: pyspiel.State object, state to search from

    Returns:
      The most visited move from the root node.
    """
    root = SearchNode(None, state.current_player(), 1)
    for _ in range(self.max_simulations):
      visit_path, working_state = self._apply_tree_policy(root, state)
      if working_state.is_terminal():
        returns = working_state.returns()
        visit_path[-1].outcome = returns
        solved = self.solve
      else:
        returns = self.evaluator.evaluate(working_state)
        solved = False

      while visit_path:
        # For chance nodes, walk up the tree to find the decision-maker.
        decision_node_idx = -1
        while visit_path[decision_node_idx].player == pyspiel.PlayerId.CHANCE:
          decision_node_idx -= 1
        # Chance node targets are for the respective decision-maker.
        target_return = returns[visit_path[decision_node_idx].player]
        node = visit_path.pop()
        node.total_reward += target_return
        node.explore_count += 1

        if solved and node.children:
          player = node.children[0].player
          if player == pyspiel.PlayerId.CHANCE:
            # Only back up chance nodes if all have the same outcome.
            # An alternative would be to back up the weighted average of
            # outcomes if all children are solved, but that is less clear.
            outcome = node.children[0].outcome
            if (outcome is not None and
                all(np.array_equal(c.outcome, outcome) for c in node.children)):
              node.outcome = outcome
            else:
              solved = False
          else:
            # If any have max utility (won?), or all children are solved,
            # choose the one best for the player choosing.
            best = None
            all_solved = True
            for child in node.children:
              if child.outcome is None:
                all_solved = False
              elif best is None or child.outcome[player] > best.outcome[player]:
                best = child
            if (best is not None and
                (all_solved or best.outcome[player] == self.max_utility)):
              node.outcome = best.outcome
            else:
              solved = False
      if root.outcome is not None:
        break

    return root
